@@ -1,5 +1,5 @@
 /* -*- C -*-
- * $Id: cptr.c 46508 2014-06-23 07:52:37Z usa $
+ * $Id: cptr.c 45820 2014-05-04 17:32:44Z nagachika $
  */
 
 #include <ruby/ruby.h>
@@ -70,7 +70,6 @@ rb_dlptr_new2(VALUE klass, void *ptr, long size, freefunc_t func)
     struct ptr_data *data;
     VALUE val;
 
-    rb_secure(4);
     val = TypedData_Make_Struct(klass, struct ptr_data, &dlptr_data_type, data);
     data->ptr = ptr;
     data->free = func;
@@ -91,7 +90,6 @@ rb_dlptr_malloc(long size, freefunc_t func)
 {
     void *ptr;
 
-    rb_secure(4);
     ptr = ruby_xmalloc((size_t)size);
     memset(ptr,0,(size_t)size);
     return rb_dlptr_new(ptr, size, func);
@@ -123,7 +121,6 @@ rb_dlptr_s_allocate(VALUE klass)
     VALUE obj;
     struct ptr_data *data;
 
-    rb_secure(4);
     obj = TypedData_Make_Struct(klass, struct ptr_data, &dlptr_data_type, data);
     data->ptr = 0;
     data->size = 0;
@@ -392,6 +389,7 @@ static VALUE
 rb_dlptr_inspect(VALUE self)
 {
     struct ptr_data *data;
+    char str[1024];
 
     TypedData_Get_Struct(self, struct ptr_data, &dlptr_data_type, data);
     return rb_sprintf("#<%"PRIsVALUE":%p ptr=%p size=%ld free=%p>",
