@@ -1,5 +1,5 @@
 /*
- * $Id: ossl_x509revoked.c 11708 2007-02-12 23:01:19Z shyouhei $
+ * $Id: ossl_x509revoked.c 31128 2011-03-19 03:30:59Z akr $
  * 'OpenSSL for Ruby' project
  * Copyright (C) 2001-2002  Michal Rokos <m.rokos@sh.cvut.cz>
  * All rights reserved.
@@ -11,20 +11,20 @@
 #include "ossl.h"
 
 #define WrapX509Rev(klass, obj, rev) do { \
-    if (!rev) { \
+    if (!(rev)) { \
 	ossl_raise(rb_eRuntimeError, "REV wasn't initialized!"); \
     } \
-    obj = Data_Wrap_Struct(klass, 0, X509_REVOKED_free, rev); \
+    (obj) = Data_Wrap_Struct((klass), 0, X509_REVOKED_free, (rev)); \
 } while (0)
 #define GetX509Rev(obj, rev) do { \
-    Data_Get_Struct(obj, X509_REVOKED, rev); \
-    if (!rev) { \
+    Data_Get_Struct((obj), X509_REVOKED, (rev)); \
+    if (!(rev)) { \
 	ossl_raise(rb_eRuntimeError, "REV wasn't initialized!"); \
     } \
 } while (0)
 #define SafeGetX509Rev(obj, rev) do { \
-    OSSL_Check_Kind(obj, cX509Rev); \
-    GetX509Rev(obj, rev); \
+    OSSL_Check_Kind((obj), cX509Rev); \
+    GetX509Rev((obj), (rev)); \
 } while (0)
 
 /*
@@ -36,7 +36,7 @@ VALUE eX509RevError;
 /*
  * PUBLIC
  */
-VALUE 
+VALUE
 ossl_x509revoked_new(X509_REVOKED *rev)
 {
     X509_REVOKED *new;
@@ -71,7 +71,7 @@ DupX509RevokedPtr(VALUE obj)
 /*
  * PRIVATE
  */
-static VALUE 
+static VALUE
 ossl_x509revoked_alloc(VALUE klass)
 {
     X509_REVOKED *rev;
@@ -85,14 +85,14 @@ ossl_x509revoked_alloc(VALUE klass)
     return obj;
 }
 
-static VALUE 
+static VALUE
 ossl_x509revoked_initialize(int argc, VALUE *argv, VALUE self)
 {
     /* EMPTY */
     return self;
 }
 
-static VALUE 
+static VALUE
 ossl_x509revoked_get_serial(VALUE self)
 {
     X509_REVOKED *rev;
@@ -102,7 +102,7 @@ ossl_x509revoked_get_serial(VALUE self)
     return asn1integer_to_num(rev->serialNumber);
 }
 
-static VALUE 
+static VALUE
 ossl_x509revoked_set_serial(VALUE self, VALUE num)
 {
     X509_REVOKED *rev;
@@ -113,17 +113,17 @@ ossl_x509revoked_set_serial(VALUE self, VALUE num)
     return num;
 }
 
-static VALUE 
+static VALUE
 ossl_x509revoked_get_time(VALUE self)
 {
     X509_REVOKED *rev;
-	
+
     GetX509Rev(self, rev);
 
     return asn1time_to_time(rev->revocationDate);
 }
 
-static VALUE 
+static VALUE
 ossl_x509revoked_set_time(VALUE self, VALUE time)
 {
     X509_REVOKED *rev;
@@ -140,7 +140,7 @@ ossl_x509revoked_set_time(VALUE self, VALUE time)
 /*
  * Gets X509v3 extensions as array of X509Ext objects
  */
-static VALUE 
+static VALUE
 ossl_x509revoked_get_extensions(VALUE self)
 {
     X509_REVOKED *rev;
@@ -166,7 +166,7 @@ ossl_x509revoked_get_extensions(VALUE self)
 /*
  * Sets X509_EXTENSIONs
  */
-static VALUE 
+static VALUE
 ossl_x509revoked_set_extensions(VALUE self, VALUE ary)
 {
     X509_REVOKED *rev;
@@ -175,14 +175,14 @@ ossl_x509revoked_set_extensions(VALUE self, VALUE ary)
     VALUE item;
 
     Check_Type(ary, T_ARRAY);
-    for (i=0; i<RARRAY(ary)->len; i++) {
-	OSSL_Check_Kind(RARRAY(ary)->ptr[i], cX509Ext);
+    for (i=0; i<RARRAY_LEN(ary); i++) {
+	OSSL_Check_Kind(RARRAY_PTR(ary)[i], cX509Ext);
     }
     GetX509Rev(self, rev);
     sk_X509_EXTENSION_pop_free(rev->extensions, X509_EXTENSION_free);
     rev->extensions = NULL;
-    for (i=0; i<RARRAY(ary)->len; i++) {
-	item = RARRAY(ary)->ptr[i];
+    for (i=0; i<RARRAY_LEN(ary); i++) {
+	item = RARRAY_PTR(ary)[i];
 	ext = DupX509ExtPtr(item);
 	if(!X509_REVOKED_add_ext(rev, ext, -1)) {
 	    ossl_raise(eX509RevError, NULL);
@@ -196,7 +196,7 @@ static VALUE
 ossl_x509revoked_add_extension(VALUE self, VALUE ext)
 {
     X509_REVOKED *rev;
-    
+
     GetX509Rev(self, rev);
     if(!X509_REVOKED_add_ext(rev, DupX509ExtPtr(ext), -1)) {
 	ossl_raise(eX509RevError, NULL);
@@ -214,10 +214,10 @@ Init_ossl_x509revoked()
     eX509RevError = rb_define_class_under(mX509, "RevokedError", eOSSLError);
 
     cX509Rev = rb_define_class_under(mX509, "Revoked", rb_cObject);
-	
+
     rb_define_alloc_func(cX509Rev, ossl_x509revoked_alloc);
     rb_define_method(cX509Rev, "initialize", ossl_x509revoked_initialize, -1);
-	
+
     rb_define_method(cX509Rev, "serial", ossl_x509revoked_get_serial, 0);
     rb_define_method(cX509Rev, "serial=", ossl_x509revoked_set_serial, 1);
     rb_define_method(cX509Rev, "time", ossl_x509revoked_get_time, 0);
