@@ -7,7 +7,8 @@
  * core routines
  */
 
-#include "ruby/ruby.h"
+#include "ruby/config.h"
+#include "ruby/defines.h"
 
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
@@ -155,7 +156,7 @@ sdbm_open(register char *file, register int flags, register int mode)
 	if (file == NULL || !*file)
 		return errno = EINVAL, (DBM *) NULL;
 /*
- * need space for two seperate filenames
+ * need space for two separate filenames
  */
 	n = strlen(file) * 2 + strlen(DIRFEXT) + strlen(PAGFEXT) + 2;
 
@@ -391,7 +392,7 @@ makroom(register DBM *db, long int hash, int need)
 		newp = (hash & db->hmask) | (db->hmask + 1);
 		debug(("newp: %ld\n", newp));
 /*
- * write delay, read avoidence/cache shuffle:
+ * write delay, read avoidance/cache shuffle:
  * select the page for incoming pair: if key is to go to the new page,
  * write out the previous one, and copy the new one over, thus making
  * it the current page. If not, simply write the new page, and we are
@@ -802,7 +803,7 @@ delpair(char *pag, datum key)
 		}
 #else
 #ifdef MEMMOVE
-		memmove(dst-m, src-m, m);
+		memmove(dst, src, m);
 #else
 		while (m--)
 			*--dst = *--src;
