@@ -1,4 +1,4 @@
-# -*- coding: euc-jp -*-
+# -*- coding: utf-8 -*-
 #
 # a dialog box with a local grab (called by 'widget')
 #
@@ -11,7 +11,7 @@ class TkDialog_Demo1 < TkDialog
   end
 
   def message
-    '⡼ܥåǤTk  "grab" ޥɤѤƥܥåǡ֥륰֡פƤޤΤ줫Υܥ¹Ԥ뤳ȤˤäơΥޤǡΥ֤ˤäƥץꥱ¾ΥɥǤϡݥ󥿴طΥ٥Ȥ뤳ȤǤʤʤäƤޤ'
+    'モーダルダイアログボックスです。Tk の "grab" コマンドを使用してダイアログボックスで「ローカルグラブ」しています。下のいずれかのボタンを実行することによって、このダイアログに答えるまで、このグラブによってアプリケーションの他のウィンドウでは、ポインタ関係のイベントを受け取ることができなくなっています。'
   end
 
   def bitmap
@@ -23,17 +23,17 @@ class TkDialog_Demo1 < TkDialog
   end
 
   def buttons
-#    "λ 󥻥 ɻ"
-    ["λ", "󥻥", "ɻ"]
+#    "了解 キャンセル コード参照"
+    ["了解", "キャンセル", "コード参照"]
   end
 end
 
 ret =  TkDialog_Demo1.new('message_config'=>{'wraplength'=>'4i'}).value
 case ret
 when 0
-  print "ʤϡλפ򲡤ޤ͡\n"
+  print "あなたは「了解」を押しましたね。\n"
 when 1
-  print "ʤϡ֥󥻥פ򲡤ޤ͡\n"
+  print "あなたは「キャンセル」を押しましたね。\n"
 when 2
   showCode 'dialog1'
 end
