@@ -2,12 +2,12 @@
 # ipaddr.rb - A class to manipulate an IP address
 #
 # Copyright (c) 2002 Hajimu UMEMOTO <ume@mahoroba.org>.
-# Copyright (c) 2007 Akinori MUSHA <knu@iDaemons.org>.
+# Copyright (c) 2007, 2009, 2012 Akinori MUSHA <knu@iDaemons.org>.
 # All rights reserved.
 #
 # You can redistribute and/or modify it under the same terms as Ruby.
 #
-# $Id: ipaddr.rb 17940 2008-07-07 08:12:23Z nobu $
+# $Id: ipaddr.rb 39300 2013-02-18 07:31:17Z knu $
 #
 # Contact:
 #   - Akinori MUSHA <knu@iDaemons.org> (current maintainer)
@@ -17,76 +17,86 @@
 #
 require 'socket'
 
-unless Socket.const_defined? "AF_INET6"
-  class Socket
-    AF_INET6 = Object.new
-  end
-
-  class << IPSocket
-    def valid_v4?(addr)
-      if /\A(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})\Z/ =~ addr
-        return $~.captures.all? {|i| i.to_i < 256}
-      end
-      return false
-    end
-
-    def valid_v6?(addr)
-      # IPv6 (normal)
-      return true if /\A[\dA-Fa-f]{1,4}(:[\dA-Fa-f]{1,4})*\Z/ =~ addr
-      return true if /\A[\dA-Fa-f]{1,4}(:[\dA-Fa-f]{1,4})*::([\dA-Fa-f]{1,4}(:[\dA-Fa-f]{1,4})*)?\Z/ =~ addr
-      return true if /\A::([\dA-Fa-f]{1,4}(:[\dA-Fa-f]{1,4})*)?\Z/ =~ addr
-      # IPv6 (IPv4 compat)
-      return true if /\A[\dA-Fa-f]{1,4}(:[\dA-Fa-f]{1,4})*:/ =~ addr && valid_v4?($')
-      return true if /\A[\dA-Fa-f]{1,4}(:[\dA-Fa-f]{1,4})*::([\dA-Fa-f]{1,4}(:[\dA-Fa-f]{1,4})*:)?/ =~ addr && valid_v4?($')
-      return true if /\A::([\dA-Fa-f]{1,4}(:[\dA-Fa-f]{1,4})*:)?/ =~ addr && valid_v4?($')
-
-      false
-    end
-
-    def valid?(addr)
-      valid_v4?(addr) || valid_v6?(addr)
-    end
-
-    alias getaddress_orig getaddress
-    def getaddress(s)
-      if valid?(s)
-        s
-      elsif /\A[-A-Za-z\d.]+\Z/ =~ s
-        getaddress_orig(s)
-      else
-        raise ArgumentError, "invalid address"
-      end
-    end
-  end
-end
-
 # IPAddr provides a set of methods to manipulate an IP address.  Both IPv4 and
 # IPv6 are supported.
 #
 # == Example
 #
 #   require 'ipaddr'
-#   
+#
 #   ipaddr1 = IPAddr.new "3ffe:505:2::1"
-#   
-#   p ipaddr1			#=> #<IPAddr: IPv6:3ffe:0505:0002:0000:0000:0000:0000:0001/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff>
-#   
-#   p ipaddr1.to_s		#=> "3ffe:505:2::1"
-#   
-#   ipaddr2 = ipaddr1.mask(48)	#=> #<IPAddr: IPv6:3ffe:0505:0002:0000:0000:0000:0000:0000/ffff:ffff:ffff:0000:0000:0000:0000:0000>
-#   
-#   p ipaddr2.to_s		#=> "3ffe:505:2::"
-#   
+#
+#   p ipaddr1                   #=> #<IPAddr: IPv6:3ffe:0505:0002:0000:0000:0000:0000:0001/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff>
+#
+#   p ipaddr1.to_s              #=> "3ffe:505:2::1"
+#
+#   ipaddr2 = ipaddr1.mask(48)  #=> #<IPAddr: IPv6:3ffe:0505:0002:0000:0000:0000:0000:0000/ffff:ffff:ffff:0000:0000:0000:0000:0000>
+#
+#   p ipaddr2.to_s              #=> "3ffe:505:2::"
+#
 #   ipaddr3 = IPAddr.new "192.168.2.0/24"
-#   
-#   p ipaddr3			#=> #<IPAddr: IPv4:192.168.2.0/255.255.255.0>
+#
+#   p ipaddr3                   #=> #<IPAddr: IPv4:192.168.2.0/255.255.255.0>
 
 class IPAddr
 
+  # 32 bit mask for IPv4
   IN4MASK = 0xffffffff
+  # 128 bit mask for IPv4
   IN6MASK = 0xffffffffffffffffffffffffffffffff
+  # Format string for IPv6
   IN6FORMAT = (["%.4x"] * 8).join(':')
 
+  # Regexp _internally_ used for parsing IPv4 address.
+  RE_IPV4ADDRLIKE = %r{
+    \A
+    (\d+) \. (\d+) \. (\d+) \. (\d+)
+    \z
+  }x
+
+  # Regexp _internally_ used for parsing IPv6 address.
+  RE_IPV6ADDRLIKE_FULL = %r{
+    \A
+    (?:
+      (?: [\da-f]{1,4} : ){7} [\da-f]{1,4}
+    |
+      ( (?: [\da-f]{1,4} : ){6} )
+      (\d+) \. (\d+) \. (\d+) \. (\d+)
+    )
+    \z
+  }xi
+
+  # Regexp _internally_ used for parsing IPv6 address.
+  RE_IPV6ADDRLIKE_COMPRESSED = %r{
+    \A
+    ( (?: (?: [\da-f]{1,4} : )* [\da-f]{1,4} )? )
+    ::
+    ( (?:
+      ( (?: [\da-f]{1,4} : )* )
+      (?:
+        [\da-f]{1,4}
+      |
+        (\d+) \. (\d+) \. (\d+) \. (\d+)
+      )
+    )? )
+    \z
+  }xi
+
+  # Generic IPAddr related error. Exceptions raised in this class should
+  # inherit from Error.
+  class Error < ArgumentError; end
+
+  # Raised when the provided IP address is an invalid address.
+  class InvalidAddressError < Error; end
+
+  # Raised when the address family is invalid such as an address with an
+  # unsupported family, an address with an inconsistent family, or an address
+  # who's family cannot be determined.
+  class AddressFamilyError < Error; end
+
+  # Raised when the address is an invalid length.
+  class InvalidPrefixError < InvalidAddressError; end
+
   # Returns the address family of this IP address.
   attr_reader :family
 
@@ -105,7 +115,7 @@ class IPAddr
     when 16
       s = IN6FORMAT % addr.unpack('n8')
     else
-      raise ArgumentError, "unsupported address family"
+      raise AddressFamilyError, "unsupported address family"
     end
     return s
   end
@@ -154,13 +164,13 @@ class IPAddr
   #   net1 = IPAddr.new("192.168.2.0/24")
   #   net2 = IPAddr.new("192.168.2.100")
   #   net3 = IPAddr.new("192.168.3.0")
-  #   p net1.include?(net2)	#=> true
-  #   p net1.include?(net3)	#=> false
+  #   p net1.include?(net2)     #=> true
+  #   p net1.include?(net3)     #=> false
   def include?(other)
     other = coerce_other(other)
     if ipv4_mapped?
       if (@mask_addr >> 32) != 0xffffffffffffffffffffffff
-	return false
+        return false
       end
       mask_addr = (@mask_addr & IN4MASK)
       addr = (@addr & IN4MASK)
@@ -197,7 +207,7 @@ class IPAddr
 
     str.gsub!(/\b0{1,3}([\da-f]+)\b/i, '\1')
     loop do
-      break if str.sub!(/\A0:0:0:0:0:0:0:0\Z/, '::')
+      break if str.sub!(/\A0:0:0:0:0:0:0:0\z/, '::')
       break if str.sub!(/\b0:0:0:0:0:0:0\b/, ':')
       break if str.sub!(/\b0:0:0:0:0:0\b/, ':')
       break if str.sub!(/\b0:0:0:0:0\b/, ':')
@@ -208,7 +218,7 @@ class IPAddr
     end
     str.sub!(/:{3,}/, '::')
 
-    if /\A::(ffff:)?([\da-f]{1,4}):([\da-f]{1,4})\Z/i =~ str
+    if /\A::(ffff:)?([\da-f]{1,4}):([\da-f]{1,4})\z/i =~ str
       str = sprintf('::%s%d.%d.%d.%d', $1, $2.hex / 256, $2.hex % 256, $3.hex / 256, $3.hex % 256)
     end
 
@@ -228,10 +238,10 @@ class IPAddr
       return [@addr].pack('N')
     when Socket::AF_INET6
       return (0..7).map { |i|
-	(@addr >> (112 - 16 * i)) & 0xffff
+        (@addr >> (112 - 16 * i)) & 0xffff
       }.pack('n8')
     else
-      raise "unsupported address family"
+      raise AddressFamilyError, "unsupported address family"
     end
   end
 
@@ -263,7 +273,7 @@ class IPAddr
   # into an IPv4-mapped IPv6 address.
   def ipv4_mapped
     if !ipv4?
-      raise ArgumentError, "not an IPv4 address"
+      raise InvalidAddressError, "not an IPv4 address"
     end
     return self.clone.set(@addr | 0xffff00000000, Socket::AF_INET6)
   end
@@ -272,7 +282,7 @@ class IPAddr
   # into an IPv4-compatible IPv6 address.
   def ipv4_compat
     if !ipv4?
-      raise ArgumentError, "not an IPv4 address"
+      raise InvalidAddressError, "not an IPv4 address"
     end
     return self.clone.set(@addr, Socket::AF_INET6)
   end
@@ -296,14 +306,14 @@ class IPAddr
     when Socket::AF_INET6
       return ip6_arpa
     else
-      raise "unsupported address family"
+      raise AddressFamilyError, "unsupported address family"
     end
   end
 
   # Returns a string for DNS reverse lookup compatible with RFC3172.
   def ip6_arpa
     if !ipv6?
-      raise ArgumentError, "not an IPv6 address"
+      raise InvalidAddressError, "not an IPv6 address"
     end
     return _reverse + ".ip6.arpa"
   end
@@ -311,7 +321,7 @@ class IPAddr
   # Returns a string for DNS reverse lookup compatible with RFC1886.
   def ip6_int
     if !ipv6?
-      raise ArgumentError, "not an IPv6 address"
+      raise InvalidAddressError, "not an IPv6 address"
     end
     return _reverse + ".ip6.int"
   end
@@ -331,6 +341,16 @@ class IPAddr
   end
   include Comparable
 
+  # Checks equality used by Hash.
+  def eql?(other)
+    return self.class == other.class && self.hash == other.hash && self == other
+  end
+
+  # Returns a hash value used by Hash, Set, and Array classes
+  def hash
+    return ([@addr, @mask_addr].hash << 1) | (ipv4? ? 0 : 1)
+  end
+
   # Creates a Range object for the network address.
   def to_range
     begin_addr = (@addr & @mask_addr)
@@ -341,7 +361,7 @@ class IPAddr
     when Socket::AF_INET6
       end_addr = (@addr | (IN6MASK ^ @mask_addr))
     else
-      raise "unsupported address family"
+      raise AddressFamilyError, "unsupported address family"
     end
 
     return clone.set(begin_addr, @family)..clone.set(end_addr, @family)
@@ -356,26 +376,29 @@ class IPAddr
     when Socket::AF_INET6
       af = "IPv6"
     else
-      raise "unsupported address family"
+      raise AddressFamilyError, "unsupported address family"
     end
     return sprintf("#<%s: %s:%s/%s>", self.class.name,
-		   af, _to_string(@addr), _to_string(@mask_addr))
+                   af, _to_string(@addr), _to_string(@mask_addr))
   end
 
   protected
 
+  # Set +@addr+, the internal stored ip address, to given +addr+. The
+  # parameter +addr+ is validated using the first +family+ member,
+  # which is +Socket::AF_INET+ or +Socket::AF_INET6+.
   def set(addr, *family)
     case family[0] ? family[0] : @family
     when Socket::AF_INET
       if addr < 0 || addr > IN4MASK
-	raise ArgumentError, "invalid address"
+        raise InvalidAddressError, "invalid address"
       end
     when Socket::AF_INET6
       if addr < 0 || addr > IN6MASK
-	raise ArgumentError, "invalid address"
+        raise InvalidAddressError, "invalid address"
       end
     else
-      raise ArgumentError, "unsupported address family"
+      raise AddressFamilyError, "unsupported address family"
     end
     @addr = addr
     if family[0]
@@ -384,18 +407,19 @@ class IPAddr
     return self
   end
 
+  # Set current netmask to given mask.
   def mask!(mask)
     if mask.kind_of?(String)
       if mask =~ /^\d+$/
-	prefixlen = mask.to_i
+        prefixlen = mask.to_i
       else
-	m = IPAddr.new(mask)
-	if m.family != @family
-	  raise ArgumentError, "address family is not same"
-	end
-	@mask_addr = m.to_i
-	@addr &= @mask_addr
-	return self
+        m = IPAddr.new(mask)
+        if m.family != @family
+          raise InvalidPrefixError, "address family is not same"
+        end
+        @mask_addr = m.to_i
+        @addr &= @mask_addr
+        return self
       end
     else
       prefixlen = mask
@@ -403,18 +427,18 @@ class IPAddr
     case @family
     when Socket::AF_INET
       if prefixlen < 0 || prefixlen > 32
-	raise ArgumentError, "invalid length"
+        raise InvalidPrefixError, "invalid length"
       end
       masklen = 32 - prefixlen
       @mask_addr = ((IN4MASK >> masklen) << masklen)
     when Socket::AF_INET6
       if prefixlen < 0 || prefixlen > 128
-	raise ArgumentError, "invalid length"
+        raise InvalidPrefixError, "invalid length"
       end
       masklen = 128 - prefixlen
       @mask_addr = ((IN6MASK >> masklen) << masklen)
     else
-      raise "unsupported address family"
+      raise AddressFamilyError, "unsupported address family"
     end
     @addr = ((@addr >> masklen) << masklen)
     return self
@@ -425,7 +449,7 @@ class IPAddr
   # Creates a new ipaddr object either from a human readable IP
   # address representation in string, or from a packed in_addr value
   # followed by an address family.
-  # 
+  #
   # In the former case, the following are the valid formats that will
   # be recognized: "address", "address/prefixlen" and "address/mask",
   # where IPv6 address may be enclosed in square brackets (`[' and
@@ -433,8 +457,8 @@ class IPAddr
   # IP address.  Although the address family is determined
   # automatically from a specified string, you can specify one
   # explicitly by the optional second argument.
-  # 
-  # Otherwise an IP addess is generated from a packed in_addr value
+  #
+  # Otherwise an IP address is generated from a packed in_addr value
   # and an address family.
   #
   # The IPAddr class defines many methods and operators, and some of
@@ -448,9 +472,9 @@ class IPAddr
         @mask_addr = (family == Socket::AF_INET) ? IN4MASK : IN6MASK
         return
       when Socket::AF_UNSPEC
-	raise ArgumentError, "address family must be specified"
+        raise AddressFamilyError, "address family must be specified"
       else
-	raise ArgumentError, "unsupported address family: #{family}"
+        raise AddressFamilyError, "unsupported address family: #{family}"
       end
     end
     prefix, prefixlen = addr.split('/')
@@ -460,17 +484,12 @@ class IPAddr
     end
     # It seems AI_NUMERICHOST doesn't do the job.
     #Socket.getaddrinfo(left, nil, Socket::AF_INET6, Socket::SOCK_STREAM, nil,
-    #		       Socket::AI_NUMERICHOST)
-    begin
-      IPSocket.getaddress(prefix)		# test if address is vaild
-    rescue
-      raise ArgumentError, "invalid address"
-    end
+    #                  Socket::AI_NUMERICHOST)
     @addr = @family = nil
     if family == Socket::AF_UNSPEC || family == Socket::AF_INET
       @addr = in_addr(prefix)
       if @addr
-	@family = Socket::AF_INET
+        @family = Socket::AF_INET
       end
     end
     if !@addr && (family == Socket::AF_UNSPEC || family == Socket::AF_INET6)
@@ -478,7 +497,7 @@ class IPAddr
       @family = Socket::AF_INET6
     end
     if family != Socket::AF_UNSPEC && @family != family
-      raise ArgumentError, "address family mismatch"
+      raise AddressFamilyError, "address family mismatch"
     end
     if prefixlen
       mask!(prefixlen)
@@ -499,26 +518,45 @@ class IPAddr
   end
 
   def in_addr(addr)
-    if addr =~ /^\d+\.\d+\.\d+\.\d+$/
-      return addr.split('.').inject(0) { |i, s|
-        i << 8 | s.to_i
-      }
+    case addr
+    when Array
+      octets = addr
+    else
+      m = RE_IPV4ADDRLIKE.match(addr) or return nil
+      octets = m.captures
     end
-    return nil
+    octets.inject(0) { |i, s|
+      (n = s.to_i) < 256 or raise InvalidAddressError, "invalid address"
+      s.match(/\A0./) and raise InvalidAddressError, "zero-filled number in IPv4 address is ambiguous"
+      i << 8 | n
+    }
   end
 
   def in6_addr(left)
     case left
-    when /^::ffff:(\d+\.\d+\.\d+\.\d+)$/i
-      return in_addr($1) + 0xffff00000000
-    when /^::(\d+\.\d+\.\d+\.\d+)$/i
-      return in_addr($1)
-    when /[^0-9a-f:]/i
-      raise ArgumentError, "invalid address"
-    when /^(.*)::(.*)$/
-      left, right = $1, $2
-    else
+    when RE_IPV6ADDRLIKE_FULL
+      if $2
+        addr = in_addr($~[2,4])
+        left = $1 + ':'
+      else
+        addr = 0
+      end
       right = ''
+    when RE_IPV6ADDRLIKE_COMPRESSED
+      if $4
+        left.count(':') <= 6 or raise InvalidAddressError, "invalid address"
+        addr = in_addr($~[4,4])
+        left = $1
+        right = $3 + '0:0'
+      else
+        left.count(':') <= ($1.empty? || $2.empty? ? 8 : 7) or
+          raise InvalidAddressError, "invalid address"
+        left = $1
+        right = $2
+        addr = 0
+      end
+    else
+      raise InvalidAddressError, "invalid address"
     end
     l = left.split(':')
     r = right.split(':')
@@ -526,9 +564,9 @@ class IPAddr
     if rest < 0
       return nil
     end
-    return (l + Array.new(rest, '0') + r).inject(0) { |i, s|
+    (l + Array.new(rest, '0') + r).inject(0) { |i, s|
       i << 16 | s.hex
-    }
+    } | addr
   end
 
   def addr_mask(addr)
@@ -538,7 +576,7 @@ class IPAddr
     when Socket::AF_INET6
       return addr & IN6MASK
     else
-      raise "unsupported address family"
+      raise AddressFamilyError, "unsupported address family"
     end
   end
 
@@ -546,12 +584,12 @@ class IPAddr
     case @family
     when Socket::AF_INET
       return (0..3).map { |i|
-	(@addr >> (8 * i)) & 0xff
+        (@addr >> (8 * i)) & 0xff
       }.join('.')
     when Socket::AF_INET6
       return ("%.32x" % @addr).reverse!.gsub!(/.(?!$)/, '\&.')
     else
-      raise "unsupported address family"
+      raise AddressFamilyError, "unsupported address family"
     end
   end
 
@@ -559,17 +597,66 @@ class IPAddr
     case @family
     when Socket::AF_INET
       return (0..3).map { |i|
-	(addr >> (24 - 8 * i)) & 0xff
+        (addr >> (24 - 8 * i)) & 0xff
       }.join('.')
     when Socket::AF_INET6
       return (("%.32x" % addr).gsub!(/.{4}(?!$)/, '\&:'))
     else
-      raise "unsupported address family"
+      raise AddressFamilyError, "unsupported address family"
     end
   end
 
 end
 
+unless Socket.const_defined? :AF_INET6
+  class Socket < BasicSocket
+    # IPv6 protocol family
+    AF_INET6 = Object.new
+  end
+
+  class << IPSocket
+    private
+
+    def valid_v6?(addr)
+      case addr
+      when IPAddr::RE_IPV6ADDRLIKE_FULL
+        if $2
+          $~[2,4].all? {|i| i.to_i < 256 }
+        else
+          true
+        end
+      when IPAddr::RE_IPV6ADDRLIKE_COMPRESSED
+        if $4
+          addr.count(':') <= 6 && $~[4,4].all? {|i| i.to_i < 256}
+        else
+          addr.count(':') <= 7
+        end
+      else
+        false
+      end
+    end
+
+    alias getaddress_orig getaddress
+
+    public
+
+    # Returns a +String+ based representation of a valid DNS hostname,
+    # IPv4 or IPv6 address.
+    #
+    #   IPSocket.getaddress 'localhost'         #=> "::1"
+    #   IPSocket.getaddress 'broadcasthost'     #=> "255.255.255.255"
+    #   IPSocket.getaddress 'www.ruby-lang.org' #=> "221.186.184.68"
+    #   IPSocket.getaddress 'www.ccc.de'        #=> "2a00:1328:e102:ccc0::122"
+    def getaddress(s)
+      if valid_v6?(s)
+        s
+      else
+        getaddress_orig(s)
+      end
+    end
+  end
+end
+
 if $0 == __FILE__
   eval DATA.read, nil, $0, __LINE__+4
 end
@@ -577,14 +664,20 @@ end
 __END__
 
 require 'test/unit'
-require 'test/unit/ui/console/testrunner'
 
 class TC_IPAddr < Test::Unit::TestCase
   def test_s_new
-    assert_nothing_raised {
-      IPAddr.new("3FFE:505:ffff::/48")
-      IPAddr.new("0:0:0:1::")
-      IPAddr.new("2001:200:300::/48")
+    [
+      ["3FFE:505:ffff::/48"],
+      ["0:0:0:1::"],
+      ["2001:200:300::/48"],
+      ["2001:200:300::192.168.1.2/48"],
+      ["1:2:3:4:5:6:7::"],
+      ["::2:3:4:5:6:7:8"],
+    ].each { |args|
+      assert_nothing_raised {
+        IPAddr.new(*args)
+      }
     }
 
     a = IPAddr.new
@@ -637,19 +730,18 @@ class TC_IPAddr < Test::Unit::TestCase
     assert_equal("2001:200:300::", IPAddr.new("2001:200:300::/48").to_s)
 
     assert_equal("2001:200:300::", IPAddr.new("[2001:200:300::]/48").to_s)
+    assert_equal("1:2:3:4:5:6:7:0", IPAddr.new("1:2:3:4:5:6:7::").to_s)
+    assert_equal("0:2:3:4:5:6:7:8", IPAddr.new("::2:3:4:5:6:7:8").to_s)
 
-    [
-      ["fe80::1%fxp0"],
-      ["::1/255.255.255.0"],
-      ["::1:192.168.1.2/120"],
-      [IPAddr.new("::1").to_i],
-      ["::ffff:192.168.1.2/120", Socket::AF_INET],
-      ["[192.168.1.2]/120"],
-    ].each { |args|
-      assert_raises(ArgumentError) {
-	IPAddr.new(*args)
-      }
-    }
+    assert_raises(IPAddr::InvalidAddressError) { IPAddr.new("192.168.0.256") }
+    assert_raises(IPAddr::InvalidAddressError) { IPAddr.new("192.168.0.011") }
+    assert_raises(IPAddr::InvalidAddressError) { IPAddr.new("fe80::1%fxp0") }
+    assert_raises(IPAddr::InvalidAddressError) { IPAddr.new("[192.168.1.2]/120") }
+    assert_raises(IPAddr::InvalidPrefixError) { IPAddr.new("::1/255.255.255.0") }
+    assert_raises(IPAddr::InvalidPrefixError) { IPAddr.new("::1/129") }
+    assert_raises(IPAddr::InvalidPrefixError) { IPAddr.new("192.168.0.1/33") }
+    assert_raises(IPAddr::AddressFamilyError) { IPAddr.new(1) }
+    assert_raises(IPAddr::AddressFamilyError) { IPAddr.new("::ffff:192.168.1.2/120", Socket::AF_INET) }
   end
 
   def test_s_new_ntoh
@@ -710,14 +802,14 @@ class TC_IPAddr < Test::Unit::TestCase
 
   def test_ip6_arpa
     assert_equal("f.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.2.0.0.0.5.0.5.0.e.f.f.3.ip6.arpa", IPAddr.new("3ffe:505:2::f").ip6_arpa)
-    assert_raises(ArgumentError) {
+    assert_raises(IPAddr::InvalidAddressError) {
       IPAddr.new("192.168.2.1").ip6_arpa
     }
   end
 
   def test_ip6_int
     assert_equal("f.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.2.0.0.0.5.0.5.0.e.f.f.3.ip6.int", IPAddr.new("3ffe:505:2::f").ip6_int)
-    assert_raises(ArgumentError) {
+    assert_raises(IPAddr::InvalidAddressError) {
       IPAddr.new("192.168.2.1").ip6_int
     }
   end
@@ -748,7 +840,7 @@ class TC_Operator < Test::Unit::TestCase
     assert_equal("3ffe:505:2:1::", a.to_s)
     assert_equal("3ffe:505:2::", @a.to_s)
     assert_equal("3ffe:505:2:1::",
-		 (@a | 0x00000000000000010000000000000000).to_s)
+                 (@a | 0x00000000000000010000000000000000).to_s)
   end
 
   def test_and
@@ -783,7 +875,9 @@ class TC_Operator < Test::Unit::TestCase
   end
 
   def test_equal
-    assert_equal(true, @a == IPAddr.new("3ffe:505:2::"))
+    assert_equal(true, @a == IPAddr.new("3FFE:505:2::"))
+    assert_equal(true, @a == IPAddr.new("3ffe:0505:0002::"))
+    assert_equal(true, @a == IPAddr.new("3ffe:0505:0002:0:0:0:0:0"))
     assert_equal(false, @a == IPAddr.new("3ffe:505:3::"))
     assert_equal(true, @a != IPAddr.new("3ffe:505:3::"))
     assert_equal(false, @a != IPAddr.new("3ffe:505:2::"))
@@ -811,4 +905,31 @@ class TC_Operator < Test::Unit::TestCase
 
   end
 
+  def test_hash
+    a1 = IPAddr.new('192.168.2.0')
+    a2 = IPAddr.new('192.168.2.0')
+    a3 = IPAddr.new('3ffe:505:2::1')
+    a4 = IPAddr.new('3ffe:505:2::1')
+    a5 = IPAddr.new('127.0.0.1')
+    a6 = IPAddr.new('::1')
+    a7 = IPAddr.new('192.168.2.0/25')
+    a8 = IPAddr.new('192.168.2.0/25')
+
+    h = { a1 => 'ipv4', a2 => 'ipv4', a3 => 'ipv6', a4 => 'ipv6', a5 => 'ipv4', a6 => 'ipv6', a7 => 'ipv4', a8 => 'ipv4'}
+    assert_equal(5, h.size)
+    assert_equal('ipv4', h[a1])
+    assert_equal('ipv4', h[a2])
+    assert_equal('ipv6', h[a3])
+    assert_equal('ipv6', h[a4])
+
+    require 'set'
+    s = Set[a1, a2, a3, a4, a5, a6, a7, a8]
+    assert_equal(5, s.size)
+    assert_equal(true, s.include?(a1))
+    assert_equal(true, s.include?(a2))
+    assert_equal(true, s.include?(a3))
+    assert_equal(true, s.include?(a4))
+    assert_equal(true, s.include?(a5))
+    assert_equal(true, s.include?(a6))
+  end
 end
