@@ -1,12 +1,10 @@
 # -*- coding: us-ascii -*-
 require "open3"
 require "timeout"
+require "test/unit"
 
 module EnvUtil
   def rubybin
-    unless ENV["RUBYOPT"]
-
-    end
     if ruby = ENV["RUBY"]
       return ruby
     end
@@ -31,20 +29,18 @@ module EnvUtil
 
   LANG_ENVS = %w"LANG LC_ALL LC_CTYPE"
 
-  def invoke_ruby(args, stdin_data="", capture_stdout=false, capture_stderr=false, opt={})
+  def invoke_ruby(args, stdin_data = "", capture_stdout = false, capture_stderr = false,
+                  encoding: nil, timeout: 10, reprieve: 1, **opt)
     in_c, in_p = IO.pipe
     out_p, out_c = IO.pipe if capture_stdout
     err_p, err_c = IO.pipe if capture_stderr && capture_stderr != :merge_to_stdout
-    opt = opt.dup
     opt[:in] = in_c
     opt[:out] = out_c if capture_stdout
     opt[:err] = capture_stderr == :merge_to_stdout ? out_c : err_c if capture_stderr
-    if enc = opt.delete(:encoding)
-      out_p.set_encoding(enc) if out_p
-      err_p.set_encoding(enc) if err_p
+    if encoding
+      out_p.set_encoding(encoding) if out_p
+      err_p.set_encoding(encoding) if err_p
     end
-    timeout = opt.delete(:timeout) || 10
-    reprieve = opt.delete(:reprieve) || 1
     c = "C"
     child_env = {}
     LANG_ENVS.each {|lc| child_env[lc] = c}
@@ -52,7 +48,7 @@ module EnvUtil
       child_env.update(args.shift)
     end
     args = [args] if args.kind_of?(String)
-    pid = spawn(child_env, EnvUtil.rubybin, *args, opt)
+    pid = spawn(child_env, EnvUtil.rubybin, *args, **opt)
     in_c.close
     out_c.close if capture_stdout
     err_c.close if capture_stderr && capture_stderr != :merge_to_stdout
@@ -61,7 +57,7 @@ module EnvUtil
     else
       th_stdout = Thread.new { out_p.read } if capture_stdout
       th_stderr = Thread.new { err_p.read } if capture_stderr && capture_stderr != :merge_to_stdout
-      in_p.write stdin_data.to_str
+      in_p.write stdin_data.to_str unless stdin_data.empty?
       in_p.close
       if (!th_stdout || th_stdout.join(timeout)) && (!th_stderr || th_stderr.join(timeout))
         stdout = th_stdout.value if capture_stdout
@@ -70,11 +66,19 @@ module EnvUtil
         signal = /mswin|mingw/ =~ RUBY_PLATFORM ? :KILL : :TERM
         begin
           Process.kill signal, pid
+          Timeout.timeout((reprieve unless signal == :KILL)) do
+            Process.wait(pid)
+          end
         rescue Errno::ESRCH
           break
+        rescue Timeout::Error
+          raise if signal == :KILL
+          signal = :KILL
         else
-        end until signal == :KILL or (sleep reprieve; signal = :KILL; false)
-        raise Timeout::Error
+          break
+        end while true
+        bt = caller_locations
+        raise Timeout::Error, "execution of #{bt.shift.label} expired", bt.map(&:to_s)
       end
       out_p.close if capture_stdout
       err_p.close if capture_stderr && capture_stderr != :merge_to_stdout
@@ -120,13 +124,65 @@ module EnvUtil
   end
   module_function :suppress_warning
 
-  def under_gc_stress
-    stress, GC.stress = GC.stress, true
+  def under_gc_stress(stress = true)
+    stress, GC.stress = GC.stress, stress
     yield
   ensure
     GC.stress = stress
   end
   module_function :under_gc_stress
+
+  def with_default_external(enc)
+    verbose, $VERBOSE = $VERBOSE, nil
+    origenc, Encoding.default_external = Encoding.default_external, enc
+    $VERBOSE = verbose
+    yield
+  ensure
+    verbose, $VERBOSE = $VERBOSE, nil
+    Encoding.default_external = origenc
+    $VERBOSE = verbose
+  end
+  module_function :with_default_external
+
+  def with_default_internal(enc)
+    verbose, $VERBOSE = $VERBOSE, nil
+    origenc, Encoding.default_internal = Encoding.default_internal, enc
+    $VERBOSE = verbose
+    yield
+  ensure
+    verbose, $VERBOSE = $VERBOSE, nil
+    Encoding.default_internal = origenc
+    $VERBOSE = verbose
+  end
+  module_function :with_default_internal
+
+  if /darwin/ =~ RUBY_PLATFORM
+    DIAGNOSTIC_REPORTS_PATH = File.expand_path("~/Library/Logs/DiagnosticReports")
+    DIAGNOSTIC_REPORTS_TIMEFORMAT = '%Y-%m-%d-%H%M%S'
+    def self.diagnostic_reports(signame, cmd, pid, now)
+      return unless %w[ABRT QUIT SEGV ILL].include?(signame)
+      cmd = File.basename(cmd)
+      path = DIAGNOSTIC_REPORTS_PATH
+      timeformat = DIAGNOSTIC_REPORTS_TIMEFORMAT
+      pat = "#{path}/#{cmd}_#{now.strftime(timeformat)}[-_]*.crash"
+      first = true
+      30.times do
+        first ? (first = false) : sleep(0.1)
+        Dir.glob(pat) do |name|
+          log = File.read(name) rescue next
+          if /\AProcess:\s+#{cmd} \[#{pid}\]$/ =~ log
+            File.unlink(name)
+            File.unlink("#{path}/.#{File.basename(name)}.plist")
+            return log
+          end
+        end
+      end
+      nil
+    end
+  else
+    def self.diagnostic_reports(signame, cmd, pid, now)
+    end
+  end
 end
 
 module Test
@@ -138,7 +194,7 @@ module Test
         code.sub!(/\A(?:\xef\xbb\xbf)?(\s*\#.*$)*(\n)?/n) {
           "#$&#{"\n" if $1 && !$2}BEGIN{throw tag, :ok}\n"
         }
-        code.force_encoding("us-ascii")
+        code.force_encoding(Encoding::UTF_8)
         verbose, $VERBOSE = $VERBOSE, nil
         yield if defined?(yield)
         case
@@ -180,24 +236,25 @@ module Test
         $VERBOSE = verbose
       end
 
-      def assert_normal_exit(testsrc, message = '', opt = {})
+      def assert_normal_exit(testsrc, message = '', child_env: nil, **opt)
         assert_valid_syntax(testsrc, caller_locations(1, 1)[0])
-        if opt.include?(:child_env)
-          opt = opt.dup
-          child_env = [opt.delete(:child_env)] || []
+        if child_env
+          child_env = [child_env]
         else
           child_env = []
         end
-        out, _, status = EnvUtil.invoke_ruby(child_env + %W'-W0', testsrc, true, :merge_to_stdout, opt)
+        out, _, status = EnvUtil.invoke_ruby(child_env + %W'-W0', testsrc, true, :merge_to_stdout, **opt)
         assert !status.signaled?, FailDesc[status, message, out]
       end
 
       FailDesc = proc do |status, message = "", out = ""|
         pid = status.pid
+        now = Time.now
         faildesc = proc do
           signo = status.termsig
-          signame = Signal.list.invert[signo]
+          signame = Signal.signame(signo)
           sigdesc = "signal #{signo}"
+          log = EnvUtil.diagnostic_reports(signame, EnvUtil.rubybin, pid, now)
           if signame
             sigdesc = "SIG#{signame} (#{sigdesc})"
           end
@@ -213,13 +270,19 @@ module Test
             full_message << "\n#{out.gsub(/^/, '| ')}"
             full_message << "\n" if /\n\z/ !~ full_message
           end
+          if log
+            full_message << "\n#{log.gsub(/^/, '| ')}"
+          end
           full_message
         end
         faildesc
       end
 
-      def assert_in_out_err(args, test_stdin = "", test_stdout = [], test_stderr = [], message = nil, opt={})
-        stdout, stderr, status = EnvUtil.invoke_ruby(args, test_stdin, true, true, opt)
+      def assert_in_out_err(args, test_stdin = "", test_stdout = [], test_stderr = [], message = nil, **opt)
+        stdout, stderr, status = EnvUtil.invoke_ruby(args, test_stdin, true, true, **opt)
+        if signo = status.termsig
+          EnvUtil.diagnostic_reports(Signal.signame(signo), EnvUtil.rubybin, status.pid, Time.now)
+        end
         if block_given?
           raise "test_stdout ignored, use block only or without block" if test_stdout != []
           raise "test_stderr ignored, use block only or without block" if test_stderr != []
@@ -243,42 +306,65 @@ module Test
         end
       end
 
-      def assert_ruby_status(args, test_stdin="", message=nil, opt={})
-        out, _, status = EnvUtil.invoke_ruby(args, test_stdin, true, :merge_to_stdout, opt)
+      def assert_ruby_status(args, test_stdin="", message=nil, **opt)
+        out, _, status = EnvUtil.invoke_ruby(args, test_stdin, true, :merge_to_stdout, **opt)
+        assert(!status.signaled?, FailDesc[status, message, out])
         message ||= "ruby exit status is not success:"
-        assert(status.success?, FailDesc[status, message, out])
+        assert(status.success?, "#{message} (#{status.inspect})")
       end
 
       ABORT_SIGNALS = Signal.list.values_at(*%w"ILL ABRT BUS SEGV")
 
-      def assert_separately(args, file = nil, line = nil, src, **opt)
+      def assert_separately(args, file = nil, line = nil, src, ignore_stderr: nil, **opt)
         unless file and line
           loc, = caller_locations(1,1)
           file ||= loc.path
           line ||= loc.lineno
         end
+        line -= 2
         src = <<eom
-  require 'test/unit';include Test::Unit::Assertions;begin;#{src}
-  ensure
+# -*- coding: #{src.encoding}; -*-
+  require #{__dir__.dump}'/envutil';include Test::Unit::Assertions
+  END {
     puts [Marshal.dump($!)].pack('m'), "assertions=\#{self._assertions}"
+  }
+#{src}
+  class Test::Unit::Runner
+    @@stop_auto_run = true
   end
 eom
-        stdout, stderr, status = EnvUtil.invoke_ruby(args, src, true, true, opt)
+        args = args.dup
+        args.insert((Hash === args.first ? 1 : 0), "--disable=gems", *$:.map {|l| "-I#{l}"})
+        stdout, stderr, status = EnvUtil.invoke_ruby(args, src, true, true, **opt)
         abort = status.coredump? || (status.signaled? && ABORT_SIGNALS.include?(status.termsig))
         assert(!abort, FailDesc[status, stderr])
         self._assertions += stdout[/^assertions=(\d+)/, 1].to_i
-        res = Marshal.load(stdout.unpack("m")[0])
-        return unless res
-        res.backtrace.each do |l|
-          l.sub!(/\A-:(\d+)/){"#{file}:#{line + $1.to_i}"}
+        begin
+          res = Marshal.load(stdout.unpack("m")[0])
+        rescue => marshal_error
+          ignore_stderr = nil
         end
-        raise res
+        if res
+          if bt = res.backtrace
+            bt.each do |l|
+              l.sub!(/\A-:(\d+)/){"#{file}:#{line + $1.to_i}"}
+            end
+          end
+          raise res
+        end
+
+        # really is it succeed?
+        unless ignore_stderr
+          # the body of assert_separately must not output anything to detect errror
+          assert_equal("", stderr, "assert_separately failed with error message")
+        end
+        assert_equal(0, status, "assert_separately failed: '#{stderr}'")
+        raise marshal_error if marshal_error
       end
 
-      def assert_warning(pat, message = nil)
+      def assert_warning(pat, msg = nil)
         stderr = EnvUtil.verbose_warning { yield }
-        message = ' "' + message + '"' if message
-        msg = proc {"warning message #{stderr.inspect} is expected to match #{pat.inspect}#{message}"}
+        msg = message(msg) {diff stderr, pat}
         assert(pat === stderr, msg)
       end
 
@@ -286,27 +372,37 @@ eom
         assert_warning(*args) {$VERBOSE = false; yield}
       end
 
-      def assert_no_memory_leak(args, prepare, code, message=nil, limit: 1.5)
+      def assert_no_memory_leak(args, prepare, code, message=nil, limit: 1.5, rss: false, **opt)
+        require_relative 'memory_status'
         token = "\e[7;1m#{$$.to_s}:#{Time.now.strftime('%s.%L')}:#{rand(0x10000).to_s(16)}:\e[m"
         token_dump = token.dump
         token_re = Regexp.quote(token)
+        envs = args.shift if Array === args and Hash === args.first
         args = [
           "--disable=gems",
           "-r", File.expand_path("../memory_status", __FILE__),
           *args,
           "-v", "-",
         ]
+        args.unshift(envs) if envs
         cmd = [
-          'END {STDERR.puts '"#{token_dump}"'"FINAL=#{Memory::Status.new.size}"}',
+          'END {STDERR.puts '"#{token_dump}"'"FINAL=#{Memory::Status.new}"}',
           prepare,
-          'STDERR.puts('"#{token_dump}"'"START=#{$initial_size = Memory::Status.new.size}")',
+          'STDERR.puts('"#{token_dump}"'"START=#{$initial_status = Memory::Status.new}")',
+          '$initial_size = $initial_status.size',
           code,
+          'GC.start',
         ].join("\n")
-        _, err, status = EnvUtil.invoke_ruby(args, cmd, true, true)
-        before = err.sub!(/^#{token_re}START=(\d+)\n/, '') && $1.to_i
-        after = err.sub!(/^#{token_re}FINAL=(\d+)\n/, '') && $1.to_i
+        _, err, status = EnvUtil.invoke_ruby(args, cmd, true, true, **opt)
+        before = err.sub!(/^#{token_re}START=(\{.*\})\n/, '') && Memory::Status.parse($1)
+        after = err.sub!(/^#{token_re}FINAL=(\{.*\})\n/, '') && Memory::Status.parse($1)
         assert_equal([true, ""], [status.success?, err], message)
-        assert_operator(after.fdiv(before), :<, limit, message)
+        ([:size, (rss && :rss)] & after.members).each do |n|
+          b = before[n]
+          a = after[n]
+          next unless a > 0 and b > 0
+          assert_operator(a.fdiv(b), :<, limit, message(message) {"#{n}: #{b} => #{a}"})
+        end
       end
 
       def assert_is_minus_zero(f)
@@ -317,7 +413,7 @@ eom
         AssertFile
       end
 
-      class << (AssertFile = Struct.new(:message).new)
+      class << (AssertFile = Struct.new(:failure_message).new)
         include Assertions
         def assert_file_predicate(predicate, *args)
           if /\Anot_/ =~ predicate
@@ -327,15 +423,15 @@ eom
           result = File.__send__(predicate, *args)
           result = !result if neg
           mesg = "Expected file " << args.shift.inspect
-          mesg << mu_pp(args) unless args.empty?
           mesg << "#{neg} to be #{predicate}"
-          mesg << " #{message}" if message
+          mesg << mu_pp(args).sub(/\A\[(.*)\]\z/m, '(\1)') unless args.empty?
+          mesg << " #{failure_message}" if failure_message
           assert(result, mesg)
         end
         alias method_missing assert_file_predicate
 
         def for(message)
-          clone.tap {|a| a.message = message}
+          clone.tap {|a| a.failure_message = message}
         end
       end
     end
