@@ -41,13 +41,14 @@ class TestArray < Test::Unit::TestCase
     assert_equal([1, 2, 3], x[1,3])
 
     x[0, 2] = 10
-    assert(x[0] == 10 && x[1] == 2)
+    assert_equal([10, 2, 3, 4, 5], x)
 
     x[0, 0] = -1
-    assert(x[0] == -1 && x[1] == 10)
+    assert_equal([-1, 10, 2, 3, 4, 5], x)
 
     x[-1, 1] = 20
-    assert(x[-1] == 20 && x.pop == 20)
+    assert_equal(20, x[-1])
+    assert_equal(20, x.pop)
   end
 
   def test_array_andor_0
@@ -97,7 +98,7 @@ class TestArray < Test::Unit::TestCase
   end
 
   def test_misc_0
-    assert(defined? "a".chomp)
+    assert(defined? "a".chomp, '"a".chomp is not defined')
     assert_equal(["a", "b", "c"], "abc".scan(/./))
     assert_equal([["1a"], ["2b"], ["3c"]], "1a2b3c".scan(/(\d.)/))
     # non-greedy match
@@ -476,20 +477,16 @@ class TestArray < Test::Unit::TestCase
 
   def test_clone
     for taint in [ false, true ]
-      for untrust in [ false, true ]
-        for frozen in [ false, true ]
-          a = @cls[*(0..99).to_a]
-          a.taint  if taint
-          a.untrust  if untrust
-          a.freeze if frozen
-          b = a.clone
+      for frozen in [ false, true ]
+        a = @cls[*(0..99).to_a]
+        a.taint  if taint
+        a.freeze if frozen
+        b = a.clone
 
-          assert_equal(a, b)
-          assert_not_equal(a.__id__, b.__id__)
-          assert_equal(a.frozen?, b.frozen?)
-          assert_equal(a.untrusted?, b.untrusted?)
-          assert_equal(a.tainted?, b.tainted?)
-        end
+        assert_equal(a, b)
+        assert_not_equal(a.__id__, b.__id__)
+        assert_equal(a.frozen?, b.frozen?)
+        assert_equal(a.tainted?, b.tainted?)
       end
     end
   end
@@ -779,10 +776,8 @@ class TestArray < Test::Unit::TestCase
 
     a6 = @cls[[1, 2], 3]
     a6.taint
-    a6.untrust
     a7 = a6.flatten
     assert_equal(true, a7.tainted?)
-    assert_equal(true, a7.untrusted?)
 
     a8 = @cls[[1, 2], 3]
     a9 = a8.flatten(0)
@@ -915,6 +910,8 @@ class TestArray < Test::Unit::TestCase
     a3 = @cls[ 'dog', 'cat' ]
     assert_equal(a1.hash, a2.hash)
     assert_not_equal(a1.hash, a3.hash)
+    bug9231 = '[ruby-core:58993] [Bug #9231]'
+    assert_not_equal(false.hash, @cls[].hash, bug9231)
   end
 
   def test_include?
@@ -973,23 +970,18 @@ class TestArray < Test::Unit::TestCase
     $, = ""
     a = @cls[1, 2, 3]
     a.taint
-    a.untrust
     s = a.join
     assert_equal(true, s.tainted?)
-    assert_equal(true, s.untrusted?)
 
     bug5902 = '[ruby-core:42161]'
-    sep = ":".taint.untrust
+    sep = ":".taint
 
     s = @cls[].join(sep)
     assert_equal(false, s.tainted?, bug5902)
-    assert_equal(false, s.untrusted?, bug5902)
     s = @cls[1].join(sep)
     assert_equal(false, s.tainted?, bug5902)
-    assert_equal(false, s.untrusted?, bug5902)
     s = @cls[1, 2].join(sep)
     assert_equal(true, s.tainted?, bug5902)
-    assert_equal(true, s.untrusted?, bug5902)
 
     e = ''.force_encoding('EUC-JP')
     u = ''.force_encoding('UTF-8')
@@ -1218,6 +1210,7 @@ class TestArray < Test::Unit::TestCase
     a = @cls[]
     i = 0
     a.reverse_each { |e|
+      i += 1
       assert(false, "Never get here")
     }
     assert_equal(0, i)
@@ -1401,6 +1394,22 @@ class TestArray < Test::Unit::TestCase
     end
   end
 
+  def test_sort_bang_with_freeze
+    ary = []
+    o1 = Object.new
+    o1.singleton_class.class_eval {
+      define_method(:<=>) {|v|
+        ary.freeze
+        1
+      }
+    }
+    o2 = o1.dup
+    ary << o1 << o2
+    orig = ary.dup
+    assert_raise(RuntimeError, "frozen during comparison") {ary.sort!}
+    assert_equal(orig, ary, "must not be modified once frozen")
+  end
+
   def test_to_a
     a = @cls[ 1, 2, 3 ]
     a_id = a.__id__
@@ -1451,6 +1460,27 @@ class TestArray < Test::Unit::TestCase
     $, = nil
   end
 
+  def test_to_h
+    kvp = Object.new
+    def kvp.to_ary
+      [:obtained, :via_to_ary]
+    end
+    array = [
+      [:key, :value],
+      kvp,
+    ]
+    assert_equal({key: :value, obtained: :via_to_ary}, array.to_h)
+
+    e = assert_raise(TypeError) {
+      [[:first_one, :ok], :not_ok].to_h
+    }
+    assert_equal "wrong element type Symbol at 1 (expected array)", e.message
+    e = assert_raise(ArgumentError) {
+      [[:first_one, :ok], [1, 2], [:not_ok]].to_h
+    }
+    assert_equal "wrong array length at 2 (expected 2, was 1)", e.message
+  end
+
   def test_uniq
     a = []
     b = a.uniq
@@ -1487,6 +1517,18 @@ class TestArray < Test::Unit::TestCase
     assert_equal(d, c)
 
     assert_equal(@cls[1, 2, 3], @cls[1, 2, 3].uniq)
+
+    a = %w(a a)
+    b = a.uniq
+    assert_equal(%w(a a), a)
+    assert(a.none?(&:frozen?))
+    assert_equal(%w(a), b)
+    assert(b.none?(&:frozen?))
+
+    bug9340 = "[ruby-core:59457]"
+    ary = [bug9340, bug9340.dup, bug9340.dup]
+    assert_equal 1, ary.uniq.size
+    assert_same bug9340, ary.uniq[0]
   end
 
   def test_uniq_with_block
@@ -1507,6 +1549,13 @@ class TestArray < Test::Unit::TestCase
     assert_equal([1,3], a)
     assert_equal([1], b)
     assert_not_same(a, b)
+
+    a = %w(a a)
+    b = a.uniq {|v| v }
+    assert_equal(%w(a a), a)
+    assert(a.none?(&:frozen?))
+    assert_equal(%w(a), b)
+    assert(b.none?(&:frozen?))
   end
 
   def test_uniq!
@@ -1553,6 +1602,13 @@ class TestArray < Test::Unit::TestCase
       a.sort_by!{|e| e[:c]}
       a.uniq!   {|e| e[:c]}
     end
+
+    a = %w(a a)
+    b = a.uniq
+    assert_equal(%w(a a), a)
+    assert(a.none?(&:frozen?))
+    assert_equal(%w(a), b)
+    assert(b.none?(&:frozen?))
   end
 
   def test_uniq_bang_with_block
@@ -1574,6 +1630,21 @@ class TestArray < Test::Unit::TestCase
     b = a.uniq! {|v| v.even? }
     assert_equal([1,2], a)
     assert_equal(nil, b)
+
+    a = %w(a a)
+    b = a.uniq! {|v| v }
+    assert_equal(%w(a), b)
+    assert_same(a, b)
+    assert b.none?(&:frozen?)
+  end
+
+  def test_uniq_bang_with_freeze
+    ary = [1,2]
+    orig = ary.dup
+    assert_raise(RuntimeError, "frozen during comparison") {
+      ary.uniq! {|v| ary.freeze; 1}
+    }
+    assert_equal(orig, ary, "must not be modified once frozen")
   end
 
   def test_unshift
@@ -1593,6 +1664,29 @@ class TestArray < Test::Unit::TestCase
     assert_equal(@cls[1,2], @cls[1] | @cls[2])
     assert_equal(@cls[1,2], @cls[1, 1] | @cls[2, 2])
     assert_equal(@cls[1,2], @cls[1, 2] | @cls[1, 2])
+
+    a = %w(a b c)
+    b = %w(a b c d e)
+    c = a | b
+    assert_equal(c, b)
+    assert_not_same(c, b)
+    assert_equal(%w(a b c), a)
+    assert_equal(%w(a b c d e), b)
+    assert(a.none?(&:frozen?))
+    assert(b.none?(&:frozen?))
+    assert(c.none?(&:frozen?))
+  end
+
+  def test_OR_in_order
+    obj1, obj2 = Class.new do
+      attr_reader :name
+      def initialize(name) @name = name; end
+      def inspect; "test_OR_in_order(#{@name})"; end
+      def hash; 0; end
+      def eql?(a) true; end
+      break [new("1"), new("2")]
+    end
+    assert_equal([obj1], [obj1]|[obj2])
   end
 
   def test_combination
@@ -1629,7 +1723,7 @@ class TestArray < Test::Unit::TestCase
 
     a = []
     [1, 2].product([0, 1, 2, 3, 4][1, 4]) {|x| a << x }
-    assert(a.all?{|x| !x.include?(0) })
+    a.all? {|x| assert_not_include(x, 0)}
   end
 
   def test_permutation
@@ -1678,7 +1772,7 @@ class TestArray < Test::Unit::TestCase
     assert_equal(@cls[1, 2, 3, 4].repeated_permutation(4).to_a, b)
 
     a = @cls[0, 1, 2, 3, 4][1, 4].repeated_permutation(2)
-    assert(a.all?{|x| !x.include?(0) })
+    assert_empty(a.reject {|x| !x.include?(0)})
   end
 
   def test_repeated_combination
@@ -1707,7 +1801,7 @@ class TestArray < Test::Unit::TestCase
     assert_equal(@cls[1, 2, 3, 4].repeated_combination(4).to_a, b)
 
     a = @cls[0, 1, 2, 3, 4][1, 4].repeated_combination(2)
-    assert(a.all?{|x| !x.include?(0) })
+    assert_empty(a.reject {|x| !x.include?(0)})
   end
 
   def test_take
@@ -1730,19 +1824,6 @@ class TestArray < Test::Unit::TestCase
     assert_equal([3,4,5,0], [1,2,3,4,5,0].drop_while {|i| i < 3 })
   end
 
-  def test_modify_check
-    a = []
-    a.freeze
-    assert_raise(RuntimeError) { a.shift }
-    a = [1, 2]
-    assert_raise(SecurityError) do
-      Thread.new do
-        $SAFE = 4
-       a.shift
-      end.value
-    end
-  end
-
   LONGP = [127, 63, 31, 15, 7].map {|x| 2**x-1 }.find do |x|
     begin
       [].first(x)
@@ -1793,6 +1874,11 @@ class TestArray < Test::Unit::TestCase
     assert_raise(ArgumentError) { [0].first(-1) }
   end
 
+  def test_last2
+    assert_equal([0], [0].last(2))
+    assert_raise(ArgumentError) { [0].last(-1) }
+  end
+
   def test_shift2
     assert_equal(0, ([0] * 16).shift)
     # check
@@ -1934,7 +2020,7 @@ class TestArray < Test::Unit::TestCase
 
     ary = Object.new
     def ary.to_a;   [1, 2]; end
-    assert_raise(TypeError, NoMethodError) {%w(a b).zip(ary)}
+    assert_raise(TypeError) {%w(a b).zip(ary)}
     def ary.each; [3, 4].each{|e|yield e}; end
     assert_equal([['a', 3], ['b', 4]], %w(a b).zip(ary))
     def ary.to_ary; [5, 6]; end
@@ -1990,13 +2076,6 @@ class TestArray < Test::Unit::TestCase
     assert_equal(A, B)
   end
 
-  def test_hash2
-    a = []
-    a << a
-    assert_equal([[a]].hash, a.hash)
-    assert_not_equal([a, a].hash, a.hash) # Implementation dependent
-  end
-
   def test_flatten_error
     a = []
     a << a
@@ -2021,6 +2100,13 @@ class TestArray < Test::Unit::TestCase
     100.times do
       assert_equal([0, 1, 2].shuffle, [0, 1, 2].shuffle(random: gen))
     end
+
+    assert_raise_with_message(ArgumentError, /unknown keyword/) do
+      [0, 1, 2].shuffle(xawqij: "a")
+    end
+    assert_raise_with_message(ArgumentError, /unknown keyword/) do
+      [0, 1, 2].shuffle!(xawqij: "a")
+    end
   end
 
   def test_shuffle_random
@@ -2094,6 +2180,10 @@ class TestArray < Test::Unit::TestCase
         assert_equal(a.sample(n), a.sample(n, random: gen), "#{i}/#{n}")
       end
     end
+
+    assert_raise_with_message(ArgumentError, /unknown keyword/) do
+      [0, 1, 2].sample(xawqij: "a")
+    end
   end
 
   def test_sample_random
@@ -2207,10 +2297,8 @@ class TestArray < Test::Unit::TestCase
   def test_inspect
     a = @cls[1, 2, 3]
     a.taint
-    a.untrust
     s = a.inspect
     assert_equal(true, s.tainted?)
-    assert_equal(true, s.untrusted?)
   end
 
   def test_initialize2
@@ -2301,8 +2389,7 @@ class TestArray < Test::Unit::TestCase
     assert_equal([], a.rotate!(13))
     assert_equal([], a.rotate!(-13))
     a = [].freeze
-    e = assert_raise(RuntimeError) {a.rotate!}
-    assert_match(/can't modify frozen/, e.message)
+    assert_raise_with_message(RuntimeError, /can't modify frozen/) {a.rotate!}
     a = [1,2,3]
     assert_raise(ArgumentError) { a.rotate!(1, 1) }
   end
