@@ -3,7 +3,6 @@ require 'test/unit'
 
 require 'tmpdir'
 require 'tempfile'
-require 'pathname'
 
 require_relative 'envutil'
 
@@ -16,7 +15,7 @@ class TestRubyOptions < Test::Unit::TestCase
 
   def with_tmpchdir
     Dir.mktmpdir {|d|
-      d = Pathname.new(d).realpath.to_s
+      d = File.realpath(d)
       Dir.chdir(d) {
         yield d
       }
@@ -187,7 +186,7 @@ class TestRubyOptions < Test::Unit::TestCase
 
     d = Dir.tmpdir
     assert_in_out_err(["-C", d, "-e", "puts Dir.pwd"]) do |r, e|
-      assert(File.identical?(r.join, d))
+      assert_file.identical?(r.join, d)
       assert_equal([], e)
     end
   end
@@ -271,20 +270,21 @@ class TestRubyOptions < Test::Unit::TestCase
     rubypath_orig = ENV['RUBYPATH']
     path_orig = ENV['PATH']
 
-    t = Tempfile.new(["test_ruby_test_rubyoption", ".rb"])
-    t.puts "p 1"
-    t.close
+    Tempfile.create(["test_ruby_test_rubyoption", ".rb"]) {|t|
+      t.puts "p 1"
+      t.close
 
-    @verbose = $VERBOSE
-    $VERBOSE = nil
+      @verbose = $VERBOSE
+      $VERBOSE = nil
 
-    ENV['PATH'] = File.dirname(t.path)
+      ENV['PATH'] = File.dirname(t.path)
 
-    assert_in_out_err(%w(-S) + [File.basename(t.path)], "", %w(1), [])
+      assert_in_out_err(%w(-S) + [File.basename(t.path)], "", %w(1), [])
 
-    ENV['RUBYPATH'] = File.dirname(t.path)
+      ENV['RUBYPATH'] = File.dirname(t.path)
 
-    assert_in_out_err(%w(-S) + [File.basename(t.path)], "", %w(1), [])
+      assert_in_out_err(%w(-S) + [File.basename(t.path)], "", %w(1), [])
+    }
 
   ensure
     if rubypath_orig
@@ -297,7 +297,6 @@ class TestRubyOptions < Test::Unit::TestCase
     else
       ENV.delete('PATH')
     end
-    t.close(true) if t
     $VERBOSE = @verbose
   end
 
@@ -333,86 +332,86 @@ class TestRubyOptions < Test::Unit::TestCase
   end
 
   def test_assignment_in_conditional
-    t = Tempfile.new(["test_ruby_test_rubyoption", ".rb"])
-    t.puts "if a = 1"
-    t.puts "end"
-    t.puts "0.times do"
-    t.puts "  if b = 2"
-    t.puts "    a += b"
-    t.puts "  end"
-    t.puts "end"
-    t.close
-    warning = ' warning: found = in conditional, should be =='
-    err = ["#{t.path}:1:#{warning}",
-           "#{t.path}:4:#{warning}",
-          ]
-    bug2136 = '[ruby-dev:39363]'
-    assert_in_out_err(["-w", t.path], "", [], err, bug2136)
-    assert_in_out_err(["-wr", t.path, "-e", ""], "", [], err, bug2136)
+    Tempfile.create(["test_ruby_test_rubyoption", ".rb"]) {|t|
+      t.puts "if a = 1"
+      t.puts "end"
+      t.puts "0.times do"
+      t.puts "  if b = 2"
+      t.puts "    a += b"
+      t.puts "  end"
+      t.puts "end"
+      t.flush
+      warning = ' warning: found = in conditional, should be =='
+      err = ["#{t.path}:1:#{warning}",
+             "#{t.path}:4:#{warning}",
+            ]
+      bug2136 = '[ruby-dev:39363]'
+      assert_in_out_err(["-w", t.path], "", [], err, bug2136)
+      assert_in_out_err(["-wr", t.path, "-e", ""], "", [], err, bug2136)
 
-    t.open
-    t.truncate(0)
-    t.puts "if a = ''; end"
-    t.puts "if a = []; end"
-    t.puts "if a = [1]; end"
-    t.puts "if a = [a]; end"
-    t.puts "if a = {}; end"
-    t.puts "if a = {1=>2}; end"
-    t.puts "if a = {3=>a}; end"
-    t.close
-    err = ["#{t.path}:1:#{warning}",
-           "#{t.path}:2:#{warning}",
-           "#{t.path}:3:#{warning}",
-           "#{t.path}:5:#{warning}",
-           "#{t.path}:6:#{warning}",
-          ]
-    feature4299 = '[ruby-dev:43083]'
-    assert_in_out_err(["-w", t.path], "", [], err, feature4299)
-    assert_in_out_err(["-wr", t.path, "-e", ""], "", [], err, feature4299)
-  ensure
-    t.close(true) if t
+      t.rewind
+      t.truncate(0)
+      t.puts "if a = ''; end"
+      t.puts "if a = []; end"
+      t.puts "if a = [1]; end"
+      t.puts "if a = [a]; end"
+      t.puts "if a = {}; end"
+      t.puts "if a = {1=>2}; end"
+      t.puts "if a = {3=>a}; end"
+      t.flush
+      err = ["#{t.path}:1:#{warning}",
+             "#{t.path}:2:#{warning}",
+             "#{t.path}:3:#{warning}",
+             "#{t.path}:5:#{warning}",
+             "#{t.path}:6:#{warning}",
+            ]
+      feature4299 = '[ruby-dev:43083]'
+      assert_in_out_err(["-w", t.path], "", [], err, feature4299)
+      assert_in_out_err(["-wr", t.path, "-e", ""], "", [], err, feature4299)
+    }
   end
 
   def test_indentation_check
-    t = Tempfile.new(["test_ruby_test_rubyoption", ".rb"])
-    t.puts "begin"
-    t.puts " end"
-    t.close
-    err = ["#{t.path}:2: warning: mismatched indentations at 'end' with 'begin' at 1"]
-    assert_in_out_err(["-w", t.path], "", [], err)
-    assert_in_out_err(["-wr", t.path, "-e", ""], "", [], err)
+    Tempfile.create(["test_ruby_test_rubyoption", ".rb"]) {|t|
+      t.puts "begin"
+      t.puts " end"
+      t.flush
+      err = ["#{t.path}:2: warning: mismatched indentations at 'end' with 'begin' at 1"]
+      assert_in_out_err(["-w", t.path], "", [], err)
+      assert_in_out_err(["-wr", t.path, "-e", ""], "", [], err)
 
-    t.open
-    t.puts "# -*- warn-indent: false -*-"
-    t.puts "begin"
-    t.puts " end"
-    t.close
-    assert_in_out_err(["-w", t.path], "", [], [], '[ruby-core:25442]')
+      t.rewind
+      t.puts "# -*- warn-indent: false -*-"
+      t.puts "begin"
+      t.puts " end"
+      t.flush
+      assert_in_out_err(["-w", t.path], "", [], [], '[ruby-core:25442]')
 
-    err = ["#{t.path}:4: warning: mismatched indentations at 'end' with 'begin' at 3"]
-    t.open
-    t.puts "# -*- warn-indent: false -*-"
-    t.puts "# -*- warn-indent: true -*-"
-    t.puts "begin"
-    t.puts " end"
-    t.close
-    assert_in_out_err(["-w", t.path], "", [], err, '[ruby-core:25442]')
+      err = ["#{t.path}:4: warning: mismatched indentations at 'end' with 'begin' at 3"]
+      t.rewind
+      t.puts "# -*- warn-indent: false -*-"
+      t.puts "# -*- warn-indent: true -*-"
+      t.puts "begin"
+      t.puts " end"
+      t.flush
+      assert_in_out_err(["-w", t.path], "", [], err, '[ruby-core:25442]')
 
-    err = ["#{t.path}:4: warning: mismatched indentations at 'end' with 'begin' at 2"]
-    t.open
-    t.puts "# -*- warn-indent: true -*-"
-    t.puts "begin"
-    t.puts "# -*- warn-indent: false -*-"
-    t.puts " end"
-    t.close
-    assert_in_out_err(["-w", t.path], "", [], [], '[ruby-core:25442]')
-  ensure
-    t.close(true) if t
+      err = ["#{t.path}:4: warning: mismatched indentations at 'end' with 'begin' at 2"]
+      t.rewind
+      t.puts "# -*- warn-indent: true -*-"
+      t.puts "begin"
+      t.puts "# -*- warn-indent: false -*-"
+      t.puts " end"
+      t.flush
+      assert_in_out_err(["-w", t.path], "", [], [], '[ruby-core:25442]')
+    }
   end
 
   def test_notfound
     notexist = "./notexist.rb"
-    rubybin = Regexp.quote(EnvUtil.rubybin)
+    rubybin = EnvUtil.rubybin.dup
+    rubybin.gsub!(%r(/), '\\') if /mswin|mingw/ =~ RUBY_PLATFORM
+    rubybin = Regexp.quote(rubybin)
     pat = Regexp.quote(notexist)
     bug1573 = '[ruby-core:23717]'
     assert_file.not_exist?(notexist)
@@ -463,17 +462,41 @@ class TestRubyOptions < Test::Unit::TestCase
     skip "platform dependent feature" if /linux|freebsd|netbsd|openbsd|darwin/ !~ RUBY_PLATFORM
 
     with_tmpchdir do
-      write_file("test-script", "$0 = 'hello world'; sleep 60")
+      write_file("test-script", "$0 = 'hello world'; /test-script/ =~ Process.argv0 or $0 = 'Process.argv0 changed!'; sleep 60")
 
       pid = spawn(EnvUtil.rubybin, "test-script")
-      sleep 0.1
-      ps = `ps -p #{pid} -o command`
+      ps = nil
+      10.times do
+        sleep 0.1
+        ps = `ps -p #{pid} -o command`
+        break if /hello world/ =~ ps
+      end
       assert_match(/hello world/, ps)
       Process.kill :KILL, pid
+      Process.wait(pid)
     end
   end
 
-  def test_segv_test
+  def test_setproctitle
+    skip "platform dependent feature" if /linux|freebsd|netbsd|openbsd|darwin/ !~ RUBY_PLATFORM
+
+    with_tmpchdir do
+      write_file("test-script", "$_0 = $0.dup; Process.setproctitle('hello world'); $0 == $_0 or Process.setproctitle('$0 changed!'); sleep 60")
+
+      pid = spawn(EnvUtil.rubybin, "test-script")
+      ps = nil
+      10.times do
+        sleep 0.1
+        ps = `ps -p #{pid} -o command`
+        break if /hello world/ =~ ps
+      end
+      assert_match(/hello world/, ps)
+      Process.kill :KILL, pid
+      Process.wait(pid)
+    end
+  end
+
+  module SEGVTest
     opts = {}
     if /mswin|mingw/ =~ RUBY_PLATFORM
       additional = '[\s\w\.\']*'
@@ -481,9 +504,11 @@ class TestRubyOptions < Test::Unit::TestCase
       opts[:rlimit_core] = 0
       additional = ""
     end
-    expected_stderr =
+    ExecOptions = opts.freeze
+
+    ExpectedStderr =
       %r(\A
-      -e:(?:1:)?\s\[BUG\]\sSegmentation\sfault\n
+      -e:(?:1:)?\s\[BUG\]\sSegmentation\sfault.*\n
       #{ Regexp.quote(RUBY_DESCRIPTION) }\n\n
       (?:--\s(?:.+\n)*\n)?
       --\sControl\sframe\sinformation\s-+\n
@@ -502,43 +527,58 @@ class TestRubyOptions < Test::Unit::TestCase
       \[NOTE\]\n
       You\smay\shave\sencountered\sa\sbug\sin\sthe\sRuby\sinterpreter\sor\sextension\slibraries.\n
       Bug\sreports\sare\swelcome.\n
-      (?:.*\n)?
       For\sdetails:\shttp:\/\/.*\.ruby-lang\.org/.*\n
       \n
       (?:#{additional})
       \z
       )x
-    assert_in_out_err(["-e", "Process.kill :SEGV, $$"], "", [], expected_stderr, nil, opts)
+  end
+
+  def test_segv_test
+    opts = SEGVTest::ExecOptions.dup
+    expected_stderr = SEGVTest::ExpectedStderr
+
+    assert_in_out_err(["--disable-gems", "-e", "Process.kill :SEGV, $$"], "", [], expected_stderr, nil, opts)
+  end
+
+  def test_segv_loaded_features
+    opts = SEGVTest::ExecOptions.dup
 
     bug7402 = '[ruby-core:49573]'
     status = assert_in_out_err(['-e', 'class Bogus; def to_str; exit true; end; end',
+                                '-e', '$".clear',
                                 '-e', '$".unshift Bogus.new',
+                                '-e', '(p $"; abort) unless $".size == 1',
                                 '-e', 'Process.kill :SEGV, $$'],
-                               "", //, /#<Bogus:/,
+                               "", [], /#<Bogus:/,
                                nil,
                                opts)
     assert_not_predicate(status, :success?, "segv but success #{bug7402}")
+  end
+
+  def test_segv_setproctitle
+    opts = SEGVTest::ExecOptions.dup
+    expected_stderr = SEGVTest::ExpectedStderr
 
     bug7597 = '[ruby-dev:46786]'
-    t = Tempfile.new(["test_ruby_test_bug7597", ".rb"])
-    t.write "f" * 100
-    t.flush
-    assert_in_out_err(["-e", "$0=ARGV[0]; Process.kill :SEGV, $$", t.path],
-                      "", [], expected_stderr, bug7597, opts)
-    t.close(true)
+    Tempfile.create(["test_ruby_test_bug7597", ".rb"]) {|t|
+      t.write "f" * 100
+      t.flush
+      assert_in_out_err(["--disable-gems", "-e", "$0=ARGV[0]; Process.kill :SEGV, $$", t.path],
+                        "", [], expected_stderr, bug7597, opts)
+    }
   end
 
   def test_DATA
-    t = Tempfile.new(["test_ruby_test_rubyoption", ".rb"])
-    t.puts "puts DATA.read.inspect"
-    t.puts "__END__"
-    t.puts "foo"
-    t.puts "bar"
-    t.puts "baz"
-    t.close
-    assert_in_out_err([t.path], "", %w("foo\\nbar\\nbaz\\n"), [])
-  ensure
-    t.close(true) if t
+    Tempfile.create(["test_ruby_test_rubyoption", ".rb"]) {|t|
+      t.puts "puts DATA.read.inspect"
+      t.puts "__END__"
+      t.puts "foo"
+      t.puts "bar"
+      t.puts "baz"
+      t.flush
+      assert_in_out_err([t.path], "", %w("foo\\nbar\\nbaz\\n"), [])
+    }
   end
 
   def test_unused_variable
