@@ -8,11 +8,17 @@ class TestSetTraceFunc < Test::Unit::TestCase
       :trace_instruction => true,
       :specialized_instruction => false
     }
+    @target_thread = Thread.current
   end
 
   def teardown
     set_trace_func(nil)
     RubyVM::InstructionSequence.compile_option = @original_compile_option
+    @target_thread = nil
+  end
+
+  def target_thread?
+    Thread.current == @target_thread
   end
 
   def test_c_call
@@ -247,14 +253,14 @@ class TestSetTraceFunc < Test::Unit::TestCase
                  events.shift)
     assert_equal(["c-return", 5, :exception, Exception],
                  events.shift)
+    assert_equal(["c-return", 5, :raise, Kernel],
+                 events.shift)
     assert_equal(["c-call", 5, :backtrace, Exception],
                  events.shift)
     assert_equal(["c-return", 5, :backtrace, Exception],
                  events.shift)
     assert_equal(["raise", 5, :test_raise, TestSetTraceFunc],
                  events.shift)
-    assert_equal(["c-return", 5, :raise, Kernel],
-                 events.shift)
     assert_equal(["c-call", 6, :===, Module],
                  events.shift)
     assert_equal(["c-return", 6, :===, Module],
@@ -397,42 +403,6 @@ class TestSetTraceFunc < Test::Unit::TestCase
     assert_equal(self, ok, bug3921)
   end
 
-  def assert_security_error_safe4(block)
-    assert_raise(SecurityError) do
-      block.call
-    end
-  end
-
-  def test_set_safe4
-    func = proc do
-      $SAFE = 4
-      set_trace_func(lambda {|*|})
-    end
-    assert_security_error_safe4(func)
-  end
-
-  def test_thread_set_safe4
-    th = Thread.start {sleep}
-    func = proc do
-      $SAFE = 4
-      th.set_trace_func(lambda {|*|})
-    end
-    assert_security_error_safe4(func)
-  ensure
-    th.kill
-  end
-
-  def test_thread_add_safe4
-    th = Thread.start {sleep}
-    func = proc do
-      $SAFE = 4
-      th.add_trace_func(lambda {|*|})
-    end
-    assert_security_error_safe4(func)
-  ensure
-    th.kill
-  end
-
   class << self
     define_method(:method_added, Module.method(:method_added))
   end
@@ -473,7 +443,7 @@ class TestSetTraceFunc < Test::Unit::TestCase
     trace = nil
     begin
     eval <<-EOF.gsub(/^.*?: /, ""), nil, 'xyzzy'
-    1: trace = TracePoint.trace(*trace_events){|tp|
+    1: trace = TracePoint.trace(*trace_events){|tp| next if !target_thread?
     2:   events << [tp.event, tp.lineno, tp.path, _defined_class.(tp), tp.method_id, tp.self, tp.binding.eval("_local_var"), _get_data.(tp)] if tp.path == 'xyzzy'
     3: }
     4: 1.times{|;_local_var| _local_var = :inner
@@ -545,10 +515,10 @@ class TestSetTraceFunc < Test::Unit::TestCase
      [:c_call,  20, "xyzzy", Exception,   :initialize,      raised_exc,  :outer, :nothing],
      [:c_return,20, "xyzzy", Exception,   :initialize,      raised_exc,  :outer, raised_exc],
      [:c_return,20, "xyzzy", Exception,   :exception,       RuntimeError, :outer, raised_exc],
+     [:c_return,20, "xyzzy", Kernel,      :raise,           self,        :outer, nil],
      [:c_call,  20, "xyzzy", Exception,   :backtrace,       raised_exc,  :outer, :nothing],
      [:c_return,20, "xyzzy", Exception,   :backtrace,       raised_exc,  :outer, nil],
      [:raise,   20, "xyzzy", TestSetTraceFunc, :trace_by_tracepoint, self, :outer, raised_exc],
-     [:c_return,20, "xyzzy", Kernel,      :raise,           self,        :outer, nil],
      [:c_call,  20, "xyzzy", Module,      :===,             RuntimeError,:outer, :nothing],
      [:c_return,20, "xyzzy", Module,      :===,             RuntimeError,:outer, true],
      [:line,    21, "xyzzy", TestSetTraceFunc, method,      self,        :outer, :nothing],
@@ -629,6 +599,7 @@ class TestSetTraceFunc < Test::Unit::TestCase
   def test_tracepoint_object_id
     tps = []
     trace = TracePoint.trace(){|tp|
+      next if !target_thread?
       tps << tp
     }
     tap{}
@@ -636,7 +607,7 @@ class TestSetTraceFunc < Test::Unit::TestCase
     tap{}
     trace.disable
 
-    # passed tp is unique, `trace' object which is genereted by TracePoint.trace
+    # passed tp is unique, `trace' object which is generated by TracePoint.trace
     tps.each{|tp|
       assert_equal trace, tp
     }
@@ -645,6 +616,7 @@ class TestSetTraceFunc < Test::Unit::TestCase
   def test_tracepoint_access_from_outside
     tp_store = nil
     trace = TracePoint.trace(){|tp|
+      next if !target_thread?
       tp_store = tp
     }
     tap{}
@@ -667,6 +639,7 @@ class TestSetTraceFunc < Test::Unit::TestCase
   def test_tracepoint_enable
     ary = []
     trace = TracePoint.new(:call){|tp|
+      next if !target_thread?
       ary << tp.method_id
     }
     foo
@@ -690,6 +663,7 @@ class TestSetTraceFunc < Test::Unit::TestCase
   def test_tracepoint_disable
     ary = []
     trace = TracePoint.trace(:call){|tp|
+      next if !target_thread?
       ary << tp.method_id
     }
     foo
@@ -730,6 +704,7 @@ class TestSetTraceFunc < Test::Unit::TestCase
 
   def test_tracepoint_return_value
     trace = TracePoint.new(:call, :return){|tp|
+      next if !target_thread?
       next if tp.path != __FILE__
       case tp.event
       when :call
@@ -750,6 +725,7 @@ class TestSetTraceFunc < Test::Unit::TestCase
 
   def test_tracepoint_raised_exception
     trace = TracePoint.new(:call, :return){|tp|
+      next if !target_thread?
       case tp.event
       when :call, :return
         assert_raise(RuntimeError) { tp.raised_exception }
@@ -775,6 +751,7 @@ class TestSetTraceFunc < Test::Unit::TestCase
   def test_tracepoint_block
     events = []
     TracePoint.new(:call, :return, :c_call, :b_call, :c_return, :b_return){|tp|
+      next if !target_thread?
       events << [
         tp.event, tp.method_id, tp.defined_class, tp.self.class,
         /return/ =~ tp.event ? tp.return_value : nil
@@ -821,6 +798,7 @@ class TestSetTraceFunc < Test::Unit::TestCase
       created_thread = Thread.new{thread_self = self}
       created_thread.join
     }
+    events.reject!{|i| i[0] != created_thread}
     assert_equal(self, thread_self)
     assert_equal([created_thread, :thread_begin, 0, nil, nil, nil, Thread], events[0])
     assert_equal([created_thread, :thread_end, 0, nil, nil, nil, Thread], events[1])
@@ -829,7 +807,10 @@ class TestSetTraceFunc < Test::Unit::TestCase
 
   def test_tracepoint_inspect
     events = []
-    trace = TracePoint.new{|tp| events << [tp.event, tp.inspect]}
+    trace = TracePoint.new{|tp|
+      next if !target_thread?
+      events << [tp.event, tp.inspect]
+    }
     assert_equal("#<TracePoint:disabled>", trace.inspect)
     trace.enable{
       assert_equal("#<TracePoint:enabled>", trace.inspect)
@@ -854,7 +835,10 @@ class TestSetTraceFunc < Test::Unit::TestCase
 
   def test_tracepoint_exception_at_line
     assert_raise(RuntimeError) do
-      TracePoint.new(:line) {raise}.enable {
+      TracePoint.new(:line) {
+        next if !target_thread?
+        raise
+      }.enable {
         1
       }
     end
@@ -897,6 +881,7 @@ class TestSetTraceFunc < Test::Unit::TestCase
   def test_trace_point_at_return_when_exception
     bug_7624 = '[ruby-core:51128] [ruby-trunk - Bug #7624]'
     TracePoint.new{|tp|
+      next if !target_thread?
       if tp.event == :return &&
         tp.method_id == :m2_test_trace_point_at_return_when_exception
         raise FOO_ERROR
@@ -910,6 +895,7 @@ class TestSetTraceFunc < Test::Unit::TestCase
     bug_7668 = '[Bug #7668]'
     ary = []
     trace = TracePoint.new{|tp|
+      next if !target_thread?
       ary << tp.event
       raise
     }
@@ -924,24 +910,6 @@ class TestSetTraceFunc < Test::Unit::TestCase
     end
   end
 
-  def test_trace_point_enable_safe4
-    tp = TracePoint.new {}
-    func = proc do
-      $SAFE = 4
-      tp.enable
-    end
-    assert_security_error_safe4(func)
-  end
-
-  def test_trace_point_disable_safe4
-    tp = TracePoint.new {}
-    func = proc do
-      $SAFE = 4
-      tp.disable
-    end
-    assert_security_error_safe4(func)
-  end
-
   def m1_for_test_trace_point_binding_in_ifunc(arg)
     arg + nil
   rescue
@@ -983,6 +951,7 @@ class TestSetTraceFunc < Test::Unit::TestCase
     # TracePoint
     tp_b = nil
     TracePoint.new(:raise) do |tp|
+      next if !target_thread?
       tp_b = tp.binding
     end.enable do
       m1_for_test_trace_point_binding_in_ifunc(0)
@@ -1007,4 +976,94 @@ class TestSetTraceFunc < Test::Unit::TestCase
       set_trace_func(nil)
     end
   end
+
+  def test_tracepoint_b_return_with_next
+    n = 0
+    TracePoint.new(:b_return){
+      next if !target_thread?
+      n += 1
+    }.enable{
+      3.times{
+        next
+      } # 3 times b_retun
+    }   # 1 time b_return
+
+    assert_equal 4, n
+  end
+
+  def test_tracepoint_b_return_with_lambda
+    n = 0
+    TracePoint.new(:b_return){
+      next if !target_thread?
+      n+=1
+    }.enable{
+      lambda{
+        return
+      }.call     # n += 1 #=> 1
+      3.times{
+        lambda{
+          return # n += 3 #=> 4
+        }.call
+      }          # n += 3 #=> 7
+      begin
+        lambda{
+          raise
+        }.call   # n += 1 #=> 8
+      rescue
+        # ignore
+      end        # n += 1 #=> 9
+    }
+
+    assert_equal 9, n
+  end
+
+  def test_isolated_raise_in_trace
+    bug9088 = '[ruby-dev:47793] [Bug #9088]'
+    assert_ruby_status([], <<-END, bug9088)
+    set_trace_func proc {raise rescue nil}
+    1.times {break}
+    END
+  end
+
+  def test_a_call
+    events = []
+    TracePoint.new(:a_call){|tp|
+      events << tp.event
+    }.enable{
+      1.times{
+        3
+      }
+      method_for_test_tracepoint_block{
+        4
+      }
+    }
+    assert_equal([
+      :b_call,
+      :c_call,
+      :b_call,
+      :call,
+      :b_call,
+    ], events)
+  end
+
+  def test_a_return
+    events = []
+    TracePoint.new(:a_return){|tp|
+      events << tp.event
+    }.enable{
+      1.times{
+        3
+      }
+      method_for_test_tracepoint_block{
+        4
+      }
+    }
+    assert_equal([
+      :b_return,
+      :c_return,
+      :b_return,
+      :return,
+      :b_return
+    ], events)
+  end
 end
