@@ -1,18 +1,11 @@
-#--
-# Copyright 2006 by Chad Fowler, Rich Kilmer, Jim Weirich and others.
-# All rights reserved.
-# See LICENSE.txt for permissions.
-#++
-
-require File.join(File.expand_path(File.dirname(__FILE__)),
-                  'gem_package_tar_test_case')
+require 'rubygems/package/tar_test_case'
 require 'rubygems/package'
 
-class TestGemPackageTarReader < TarTestCase
+class TestGemPackageTarReader < Gem::Package::TarTestCase
 
   def test_each_entry
-    tar = tar_dir_header "foo", "bar", 0
-    tar << tar_file_header("bar", "baz", 0, 0)
+    tar = tar_dir_header "foo", "bar", 0, Time.now
+    tar << tar_file_header("bar", "baz", 0, 0, Time.now)
 
     io = TempIO.new tar
 
@@ -32,8 +25,9 @@ class TestGemPackageTarReader < TarTestCase
   def test_rewind
     content = ('a'..'z').to_a.join(" ")
 
-    str = tar_file_header("lib/foo", "", 010644, content.size) + content +
-            "\0" * (512 - content.size)
+    str =
+      tar_file_header("lib/foo", "", 010644, content.size, Time.now) +
+        content + "\0" * (512 - content.size)
     str << "\0" * 1024
 
     Gem::Package::TarReader.new(TempIO.new(str)) do |tar_reader|
@@ -49,5 +43,37 @@ class TestGemPackageTarReader < TarTestCase
     end
   end
 
+  def test_seek
+    tar = tar_dir_header "foo", "bar", 0, Time.now
+    tar << tar_file_header("bar", "baz", 0, 0, Time.now)
+
+    io = TempIO.new tar
+
+    Gem::Package::TarReader.new io do |tar_reader|
+      tar_reader.seek 'baz/bar' do |entry|
+        assert_kind_of Gem::Package::TarReader::Entry, entry
+
+        assert_equal 'baz/bar', entry.full_name
+      end
+
+      assert_equal 0, io.pos
+    end
+  end
+
+  def test_seek_missing
+    tar = tar_dir_header "foo", "bar", 0, Time.now
+    tar << tar_file_header("bar", "baz", 0, 0, Time.now)
+
+    io = TempIO.new tar
+
+    Gem::Package::TarReader.new io do |tar_reader|
+      tar_reader.seek 'nonexistent' do |entry|
+        flunk 'entry missing but entry-found block was run'
+      end
+
+      assert_equal 0, io.pos
+    end
+  end
+
 end
 
