@@ -23,6 +23,7 @@ class TestSocket_UNIXSocket < Test::Unit::TestCase
       r2 = s2.recv_io
       assert_equal(r1.stat.ino, r2.stat.ino)
       assert_not_equal(r1.fileno, r2.fileno)
+      assert(r2.close_on_exec?)
       w.syswrite "a"
       assert_equal("a", r2.sysread(10))
     ensure
@@ -62,6 +63,7 @@ class TestSocket_UNIXSocket < Test::Unit::TestCase
         send_io_ary.length.times {|i|
           assert_not_equal(send_io_ary[i].fileno, recv_io_ary[i].fileno)
           assert(File.identical?(send_io_ary[i], recv_io_ary[i]))
+          assert(recv_io_ary[i].close_on_exec?)
         }
       }
     }
@@ -98,6 +100,7 @@ class TestSocket_UNIXSocket < Test::Unit::TestCase
         send_io_ary.length.times {|i|
           assert_not_equal(send_io_ary[i].fileno, recv_io_ary[i].fileno)
           assert(File.identical?(send_io_ary[i], recv_io_ary[i]))
+          assert(recv_io_ary[i].close_on_exec?)
         }
       }
     }
@@ -151,6 +154,7 @@ class TestSocket_UNIXSocket < Test::Unit::TestCase
         r2 = s2.recv_io
         begin
           assert(File.identical?(r1, r2))
+          assert(r2.close_on_exec?)
         ensure
           r2.close
         end
@@ -231,6 +235,7 @@ class TestSocket_UNIXSocket < Test::Unit::TestCase
 	  r2 = ios[0]
 	  begin
 	    assert(File.identical?(r1, r2))
+            assert(r2.close_on_exec?)
 	  ensure
 	    r2.close
 	  end
@@ -240,7 +245,7 @@ class TestSocket_UNIXSocket < Test::Unit::TestCase
   end
 
   def bound_unix_socket(klass)
-    tmpfile = Tempfile.new("testrubysock")
+    tmpfile = Tempfile.new("s")
     path = tmpfile.path
     tmpfile.close(true)
     yield klass.new(path), path
@@ -261,6 +266,16 @@ class TestSocket_UNIXSocket < Test::Unit::TestCase
     }
   end
 
+  def test_cloexec
+    bound_unix_socket(UNIXServer) {|serv, path|
+      c = UNIXSocket.new(path)
+      s = serv.accept
+      assert(serv.close_on_exec?)
+      assert(c.close_on_exec?)
+      assert(s.close_on_exec?)
+    }
+  end
+
   def test_noname_path
     s1, s2 = UNIXSocket.pair
     assert_equal("", s1.path)
@@ -334,7 +349,13 @@ class TestSocket_UNIXSocket < Test::Unit::TestCase
 
   def test_dgram_pair
     s1, s2 = UNIXSocket.pair(Socket::SOCK_DGRAM)
-    assert_raise(Errno::EAGAIN) { s1.recv_nonblock(10) }
+    begin
+      s1.recv_nonblock(10)
+      fail
+    rescue => e
+      assert(IO::EAGAINWaitReadable === e)
+      assert(IO::WaitReadable === e)
+    end
     s2.send("", 0)
     s2.send("haha", 0)
     s2.send("", 0)
@@ -343,7 +364,7 @@ class TestSocket_UNIXSocket < Test::Unit::TestCase
     assert_equal("haha", s1.recv(10))
     assert_equal("", s1.recv(10))
     assert_equal("", s1.recv(10))
-    assert_raise(Errno::EAGAIN) { s1.recv_nonblock(10) }
+    assert_raise(IO::EAGAINWaitReadable) { s1.recv_nonblock(10) }
   ensure
     s1.close if s1
     s2.close if s2
@@ -358,7 +379,7 @@ class TestSocket_UNIXSocket < Test::Unit::TestCase
         pipe[0].read_nonblock(1)
         fail
       rescue => e
-        assert(IO::WaitReadable === e)
+        assert(IO::EAGAINWaitReadable === e)
       end
     end
     Timeout.timeout(10) do
@@ -400,6 +421,14 @@ class TestSocket_UNIXSocket < Test::Unit::TestCase
     assert_kind_of(UNIXSocket, pair[1])
   end
 
+  def test_unix_socket_pair_close_on_exec
+    pair = nil
+    UNIXSocket.pair {|s1, s2|
+      assert(s1.close_on_exec?)
+      assert(s2.close_on_exec?)
+    }
+  end
+
   def test_initialize
     Dir.mktmpdir {|d|
       Socket.open(Socket::AF_UNIX, Socket::SOCK_STREAM, 0) {|s|
