@@ -89,6 +89,14 @@ class TestStringIO < Test::Unit::TestCase
     f.close unless f.closed?
   end
 
+  def test_write_nonblock_no_exceptions
+    s = ""
+    f = StringIO.new(s, "w")
+    f.write_nonblock("foo", exception: false)
+    f.close
+    assert_equal("foo", s)
+  end
+
   def test_write_nonblock
     s = ""
     f = StringIO.new(s, "w")
@@ -111,24 +119,6 @@ class TestStringIO < Test::Unit::TestCase
     f.close unless f.closed?
   end
 
-  def test_write_infection
-    bug9769 = '[ruby-dev:48118] [Bug #9769]'
-    s = "".untaint
-    f = StringIO.new(s, "w")
-    f.print("bar".taint)
-    f.close
-    assert_predicate(s, :tainted?, bug9769)
-  ensure
-    f.close unless f.closed?
-  end
-
-  def test_write_encoding
-    s = "".force_encoding(Encoding::UTF_8)
-    f = StringIO.new(s)
-    f.print("\u{3053 3093 306b 3061 306f ff01}".b)
-    assert_equal(Encoding::UTF_8, s.encoding, "honor the original encoding over ASCII-8BIT")
-  end
-
   def test_mode_error
     f = StringIO.new("", "r")
     assert_raise(IOError) { f.write("foo") }
@@ -429,22 +419,6 @@ class TestStringIO < Test::Unit::TestCase
     assert_equal("foo123", s)
   end
 
-  def test_putc_nonascii
-    s = ""
-    f = StringIO.new(s, "w")
-    f.putc("\u{3042}")
-    f.putc(0x3044)
-    f.close
-    assert_equal("\u{3042}D", s)
-
-    s = "foo"
-    f = StringIO.new(s, "a")
-    f.putc("\u{3042}")
-    f.putc(0x3044)
-    f.close
-    assert_equal("foo\u{3042}D", s)
-  end
-
   def test_read
     f = StringIO.new("\u3042\u3044")
     assert_raise(ArgumentError) { f.read(-1) }
@@ -471,7 +445,7 @@ class TestStringIO < Test::Unit::TestCase
     f = StringIO.new("\u3042\u3044")
     assert_raise(ArgumentError) { f.readpartial(-1) }
     assert_raise(ArgumentError) { f.readpartial(1, 2, 3) }
-    assert_equal("\u3042\u3044", f.readpartial)
+    assert_equal("\u3042\u3044".force_encoding(Encoding::ASCII_8BIT), f.readpartial(100))
     f.rewind
     assert_equal("\u3042\u3044".force_encoding(Encoding::ASCII_8BIT), f.readpartial(f.size))
     f.rewind
@@ -484,7 +458,19 @@ class TestStringIO < Test::Unit::TestCase
     f = StringIO.new("\u3042\u3044")
     assert_raise(ArgumentError) { f.read_nonblock(-1) }
     assert_raise(ArgumentError) { f.read_nonblock(1, 2, 3) }
-    assert_equal("\u3042\u3044", f.read_nonblock)
+    assert_equal("\u3042\u3044".force_encoding("BINARY"), f.read_nonblock(100))
+    assert_raise(EOFError) { f.read_nonblock(10) }
+    f.rewind
+    assert_equal("\u3042\u3044".force_encoding(Encoding::ASCII_8BIT), f.read_nonblock(f.size))
+  end
+
+  def test_read_nonblock_no_exceptions
+    f = StringIO.new("\u3042\u3044")
+    assert_raise(ArgumentError) { f.read_nonblock(-1, exception: false) }
+    assert_raise(ArgumentError) { f.read_nonblock(1, 2, 3, exception: false) }
+    assert_raise(ArgumentError) { f.read_nonblock }
+    assert_equal("\u3042\u3044".force_encoding(Encoding::ASCII_8BIT), f.read_nonblock(100, exception: false))
+    assert_equal(nil, f.read_nonblock(10, exception: false))
     f.rewind
     assert_equal("\u3042\u3044".force_encoding(Encoding::ASCII_8BIT), f.read_nonblock(f.size))
     f.rewind
