@@ -2,9 +2,16 @@
  * This file is included by vm.c
  */
 
-#define CACHE_SIZE 0x800
-#define CACHE_MASK 0x7ff
-#define EXPR1(c,m) ((((c)>>3)^(m))&CACHE_MASK)
+#ifndef GLOBAL_METHOD_CACHE_SIZE
+#define GLOBAL_METHOD_CACHE_SIZE 0x800
+#endif
+#ifndef GLOBAL_METHOD_CACHE_MASK
+#define GLOBAL_METHOD_CACHE_MASK 0x7ff
+#endif
+
+#define GLOBAL_METHOD_CACHE_KEY(c,m) ((((c)>>3)^(m))&GLOBAL_METHOD_CACHE_MASK)
+#define GLOBAL_METHOD_CACHE(c,m) (global_method_cache + GLOBAL_METHOD_CACHE_KEY(c,m))
+#include "method.h"
 
 #define NOEX_NOREDEF 0
 #ifndef NOEX_NOREDEF
@@ -13,57 +20,65 @@
 
 static void rb_vm_check_redefinition_opt_method(const rb_method_entry_t *me, VALUE klass);
 
-static ID object_id;
-static ID removed, singleton_removed, undefined, singleton_undefined;
-static ID added, singleton_added, attached;
+#define object_id           idObject_id
+#define added               idMethod_added
+#define singleton_added     idSingleton_method_added
+#define removed             idMethod_removed
+#define singleton_removed   idSingleton_method_removed
+#define undefined           idMethod_undefined
+#define singleton_undefined idSingleton_method_undefined
+#define attached            id__attached__
 
-struct cache_entry {		/* method hash table. */
-    VALUE filled_version;        /* filled state version */
-    ID mid;			/* method's id */
-    VALUE klass;		/* receiver's class */
-    rb_method_entry_t *me;
+struct cache_entry {
+    rb_serial_t method_state;
+    rb_serial_t class_serial;
+    ID mid;
+    rb_method_entry_t* me;
     VALUE defined_class;
 };
 
-static struct cache_entry cache[CACHE_SIZE];
+static struct cache_entry global_method_cache[GLOBAL_METHOD_CACHE_SIZE];
 #define ruby_running (GET_VM()->running)
 /* int ruby_running = 0; */
 
 static void
-vm_clear_global_method_cache(void)
+rb_class_clear_method_cache(VALUE klass)
 {
-    struct cache_entry *ent, *end;
-
-    ent = cache;
-    end = ent + CACHE_SIZE;
-    while (ent < end) {
-	ent->filled_version = 0;
-	ent++;
-    }
+    RCLASS_SERIAL(klass) = rb_next_class_serial();
+    rb_class_foreach_subclass(klass, rb_class_clear_method_cache);
 }
 
 void
 rb_clear_cache(void)
 {
-    rb_vm_change_state();
-}
-
-static void
-rb_clear_cache_for_undef(VALUE klass, ID id)
-{
-    rb_vm_change_state();
-}
-
-static void
-rb_clear_cache_by_id(ID id)
-{
-    rb_vm_change_state();
+    rb_warning("rb_clear_cache() is deprecated.");
+    INC_GLOBAL_METHOD_STATE();
+    INC_GLOBAL_CONSTANT_STATE();
 }
 
 void
-rb_clear_cache_by_class(VALUE klass)
+rb_clear_constant_cache(void)
 {
-    rb_vm_change_state();
+    INC_GLOBAL_CONSTANT_STATE();
+}
+
+void
+rb_clear_method_cache_by_class(VALUE klass)
+{
+    if (klass && klass != Qundef) {
+	int global = klass == rb_cBasicObject || klass == rb_cObject || klass == rb_mKernel;
+
+	if (RUBY_DTRACE_METHOD_CACHE_CLEAR_ENABLED()) {
+	    RUBY_DTRACE_METHOD_CACHE_CLEAR(global ? "global" : rb_class2name(klass), rb_sourcefile(), rb_sourceline());
+	}
+
+	if (global) {
+	    INC_GLOBAL_METHOD_STATE();
+	}
+	else {
+	    rb_class_clear_method_cache(klass);
+	}
+    }
 }
 
 VALUE
@@ -122,26 +137,20 @@ void
 rb_sweep_method_entry(void *pvm)
 {
     rb_vm_t *vm = pvm;
-    struct unlinked_method_entry_list_entry *ume = vm->unlinked_method_entry_list, *prev_ume = 0, *curr_ume;
+    struct unlinked_method_entry_list_entry **prev_ume = &vm->unlinked_method_entry_list, *ume = *prev_ume, *curr_ume;
 
     while (ume) {
 	if (ume->me->mark) {
 	    ume->me->mark = 0;
-	    prev_ume = ume;
-	    ume = ume->next;
+	    prev_ume = &ume->next;
+	    ume = *prev_ume;
 	}
 	else {
 	    rb_free_method_entry(ume->me);
 
-	    if (prev_ume == 0) {
-		vm->unlinked_method_entry_list = ume->next;
-	    }
-	    else {
-		prev_ume->next = ume->next;
-	    }
-
 	    curr_ume = ume;
 	    ume = ume->next;
+	    *prev_ume = ume;
 	    xfree(curr_ume);
 	}
     }
@@ -172,7 +181,6 @@ rb_free_method_entry(rb_method_entry_t *me)
     xfree(me);
 }
 
-static inline rb_method_entry_t *search_method(VALUE klass, ID id, VALUE *defined_class_ptr);
 static int rb_method_definition_eq(const rb_method_definition_t *d1, const rb_method_definition_t *d2);
 
 static inline rb_method_entry_t *
@@ -215,6 +223,7 @@ rb_add_refined_method_entry(VALUE refined_class, ID mid)
 
     if (me) {
 	make_method_entry_refined(me);
+	rb_clear_method_cache_by_class(refined_class);
     }
     else {
 	rb_add_method(refined_class, mid, VM_METHOD_TYPE_REFINED, 0,
@@ -238,10 +247,6 @@ rb_method_entry_make(VALUE klass, ID mid, rb_method_type_t type,
     if (NIL_P(klass)) {
 	klass = rb_cObject;
     }
-    if (rb_safe_level() >= 4 &&
-	(klass == rb_cObject || !OBJ_UNTRUSTED(klass))) {
-	rb_raise(rb_eSecurityError, "Insecure: can't define method");
-    }
     if (!FL_TEST(klass, FL_SINGLETON) &&
 	type != VM_METHOD_TYPE_NOTIMPLEMENTED &&
 	type != VM_METHOD_TYPE_ZSUPER &&
@@ -306,7 +311,7 @@ rb_method_entry_make(VALUE klass, ID mid, rb_method_type_t type,
 		break;
 	    }
 	    if (iseq && !NIL_P(iseq->location.path)) {
-		int line = iseq->line_info_table ? rb_iseq_first_lineno(iseq) : 0;
+		int line = iseq->line_info_table ? FIX2INT(rb_iseq_first_lineno(iseq->self)) : 0;
 		rb_compile_warning(RSTRING_PTR(iseq->location.path), line,
 				   "previous definition of %s was here",
 				   rb_id2name(old_def->original_id));
@@ -318,14 +323,31 @@ rb_method_entry_make(VALUE klass, ID mid, rb_method_type_t type,
 
     me = ALLOC(rb_method_entry_t);
 
-    rb_clear_cache_by_id(mid);
+    rb_clear_method_cache_by_class(klass);
 
     me->flag = NOEX_WITH_SAFE(noex);
     me->mark = 0;
     me->called_id = mid;
-    me->klass = defined_class;
+    RB_OBJ_WRITE(klass, &me->klass, defined_class);
     me->def = def;
-    if (def) def->alias_count++;
+
+    if (def) {
+	def->alias_count++;
+
+	switch(def->type) {
+	  case VM_METHOD_TYPE_ISEQ:
+	    RB_OBJ_WRITTEN(klass, Qundef, def->body.iseq->self);
+	    break;
+	  case VM_METHOD_TYPE_IVAR:
+	    RB_OBJ_WRITTEN(klass, Qundef, def->body.attr.location);
+	    break;
+	  case VM_METHOD_TYPE_BMETHOD:
+	    RB_OBJ_WRITTEN(klass, Qundef, def->body.proc);
+	    break;
+	  default:;
+	    /* ignore */
+	}
+    }
 
     /* check mid */
     if (klass == rb_cObject && mid == idInitialize) {
@@ -333,7 +355,7 @@ rb_method_entry_make(VALUE klass, ID mid, rb_method_type_t type,
     }
     /* check mid */
     if (mid == object_id || mid == id__send__) {
-	if (type == VM_METHOD_TYPE_ISEQ && search_method(klass, mid, 0)) {
+	if (type == VM_METHOD_TYPE_ISEQ) {
 	    rb_warn("redefining `%s' may cause serious problems", rb_id2name(mid));
 	}
     }
@@ -419,9 +441,12 @@ rb_add_method(VALUE klass, ID mid, rb_method_type_t type, void *opts, rb_method_
     def->original_id = mid;
     def->alias_count = 0;
     switch (type) {
-      case VM_METHOD_TYPE_ISEQ:
-	def->body.iseq = (rb_iseq_t *)opts;
-	break;
+      case VM_METHOD_TYPE_ISEQ: {
+	  rb_iseq_t *iseq = (rb_iseq_t *)opts;
+	  *(rb_iseq_t **)&def->body.iseq = iseq;
+	  RB_OBJ_WRITTEN(klass, Qundef, iseq->self);
+	  break;
+      }
       case VM_METHOD_TYPE_CFUNC:
 	{
 	    rb_method_cfunc_t *cfunc = (rb_method_cfunc_t *)opts;
@@ -431,16 +456,16 @@ rb_add_method(VALUE klass, ID mid, rb_method_type_t type, void *opts, rb_method_
       case VM_METHOD_TYPE_ATTRSET:
       case VM_METHOD_TYPE_IVAR:
 	def->body.attr.id = (ID)opts;
-	def->body.attr.location = Qfalse;
+	RB_OBJ_WRITE(klass, &def->body.attr.location, Qfalse);
 	th = GET_THREAD();
 	cfp = rb_vm_get_ruby_level_next_cfp(th, th->cfp);
 	if (cfp && (line = rb_vm_get_sourceline(cfp))) {
 	    VALUE location = rb_ary_new3(2, cfp->iseq->location.path, INT2FIX(line));
-	    def->body.attr.location = rb_ary_freeze(location);
+	    RB_OBJ_WRITE(klass, &def->body.attr.location, rb_ary_freeze(location));
 	}
 	break;
       case VM_METHOD_TYPE_BMETHOD:
-	def->body.proc = (VALUE)opts;
+	RB_OBJ_WRITE(klass, &def->body.proc, (VALUE)opts);
 	break;
       case VM_METHOD_TYPE_NOTIMPLEMENTED:
 	setup_method_cfunc_struct(&def->body.cfunc, rb_f_notimplement, -1);
@@ -522,6 +547,12 @@ search_method(VALUE klass, ID id, VALUE *defined_class_ptr)
     return me;
 }
 
+rb_method_entry_t *
+rb_method_entry_at(VALUE klass, ID id)
+{
+    return lookup_method_table(klass, id);
+}
+
 /*
  * search method entry without the method cache.
  *
@@ -535,30 +566,22 @@ rb_method_entry_get_without_cache(VALUE klass, ID id,
     VALUE defined_class;
     rb_method_entry_t *me = search_method(klass, id, &defined_class);
 
-    if (me && me->klass) {
-	switch (BUILTIN_TYPE(me->klass)) {
-	  case T_CLASS:
-	    if (RBASIC(klass)->flags & FL_SINGLETON) break;
-	    /* fall through */
-	  case T_ICLASS:
-	    defined_class = me->klass;
-	}
-    }
+    if (me && RB_TYPE_P(me->klass, T_ICLASS))
+	defined_class = me->klass;
 
     if (ruby_running) {
 	struct cache_entry *ent;
-	ent = cache + EXPR1(klass, id);
-	ent->filled_version = GET_VM_STATE_VERSION();
-	ent->klass = klass;
+	ent = GLOBAL_METHOD_CACHE(klass, id);
+	ent->class_serial = RCLASS_EXT(klass)->class_serial;
+	ent->method_state = GET_GLOBAL_METHOD_STATE();
 	ent->defined_class = defined_class;
+	ent->mid = id;
 
 	if (UNDEFINED_METHOD_ENTRY_P(me)) {
-	    ent->mid = id;
 	    ent->me = 0;
 	    me = 0;
 	}
 	else {
-	    ent->mid = id;
 	    ent->me = me;
 	}
     }
@@ -568,17 +591,34 @@ rb_method_entry_get_without_cache(VALUE klass, ID id,
     return me;
 }
 
+#if VM_DEBUG_VERIFY_METHOD_CACHE
+static void
+verify_method_cache(VALUE klass, ID id, VALUE defined_class, rb_method_entry_t *me)
+{
+    VALUE actual_defined_class;
+    rb_method_entry_t *actual_me =
+	rb_method_entry_get_without_cache(klass, id, &actual_defined_class);
+
+    if (me != actual_me || defined_class != actual_defined_class) {
+	rb_bug("method cache verification failed");
+    }
+}
+#endif
+
 rb_method_entry_t *
 rb_method_entry(VALUE klass, ID id, VALUE *defined_class_ptr)
 {
 #if OPT_GLOBAL_METHOD_CACHE
     struct cache_entry *ent;
-
-    ent = cache + EXPR1(klass, id);
-    if (ent->filled_version == GET_VM_STATE_VERSION() &&
-	ent->mid == id && ent->klass == klass) {
+    ent = GLOBAL_METHOD_CACHE(klass, id);
+    if (ent->method_state == GET_GLOBAL_METHOD_STATE() &&
+	ent->class_serial == RCLASS_EXT(klass)->class_serial &&
+	ent->mid == id) {
 	if (defined_class_ptr)
 	    *defined_class_ptr = ent->defined_class;
+#if VM_DEBUG_VERIFY_METHOD_CACHE
+	verify_method_cache(klass, id, ent->defined_class, ent->me);
+#endif
 	return ent->me;
     }
 #endif
@@ -677,12 +717,6 @@ remove_method(VALUE klass, ID mid)
     VALUE self = klass;
 
     klass = RCLASS_ORIGIN(klass);
-    if (klass == rb_cObject) {
-	rb_secure(4);
-    }
-    if (rb_safe_level() >= 4 && !OBJ_UNTRUSTED(klass)) {
-	rb_raise(rb_eSecurityError, "Insecure: can't remove method");
-    }
     rb_check_frozen(klass);
     if (mid == object_id || mid == id__send__ || mid == idInitialize) {
 	rb_warn("removing `%s' may cause serious problems", rb_id2name(mid));
@@ -690,23 +724,17 @@ remove_method(VALUE klass, ID mid)
 
     if (!st_lookup(RCLASS_M_TBL(klass), mid, &data) ||
 	!(me = (rb_method_entry_t *)data) ||
-	(!me->def || me->def->type == VM_METHOD_TYPE_UNDEF) ||
-        UNDEFINED_REFINED_METHOD_P(me->def)) {
+	(!me->def || me->def->type == VM_METHOD_TYPE_UNDEF)) {
 	rb_name_error(mid, "method `%s' not defined in %s",
 		      rb_id2name(mid), rb_class2name(klass));
     }
-
     key = (st_data_t)mid;
     st_delete(RCLASS_M_TBL(klass), &key, &data);
 
     rb_vm_check_redefinition_opt_method(me, klass);
-    rb_clear_cache_for_undef(klass, mid);
+    rb_clear_method_cache_by_class(klass);
     rb_unlink_method_entry(me);
 
-    if (me->def->type == VM_METHOD_TYPE_REFINED) {
-	rb_add_refined_method_entry(klass, mid);
-    }
-
     CALL_METHOD_HOOK(self, removed, mid);
 }
 
@@ -725,9 +753,11 @@ rb_remove_method(VALUE klass, const char *name)
 /*
  *  call-seq:
  *     remove_method(symbol)   -> self
+ *     remove_method(string)   -> self
  *
  *  Removes the method identified by _symbol_ from the current
  *  class. For an example, see <code>Module.undef_method</code>.
+ *  String arguments are converted to symbols.
  */
 
 static VALUE
@@ -768,17 +798,12 @@ rb_export_method(VALUE klass, ID name, rb_method_flag_t noex)
     rb_method_entry_t *me;
     VALUE defined_class;
 
-    if (klass == rb_cObject) {
-	rb_secure(4);
-    }
-
     me = search_method(klass, name, &defined_class);
     if (!me && RB_TYPE_P(klass, T_MODULE)) {
 	me = search_method(rb_cObject, name, &defined_class);
     }
 
-    if (UNDEFINED_METHOD_ENTRY_P(me) ||
-	UNDEFINED_REFINED_METHOD_P(me->def)) {
+    if (UNDEFINED_METHOD_ENTRY_P(me)) {
 	rb_print_undef(klass, name, 0);
     }
 
@@ -791,6 +816,7 @@ rb_export_method(VALUE klass, ID name, rb_method_flag_t noex)
 	    if (me->def->type == VM_METHOD_TYPE_REFINED) {
 		me->def->body.orig_me->flag = noex;
 	    }
+	    rb_clear_method_cache_by_class(klass);
 	}
 	else {
 	    rb_add_method(klass, name, VM_METHOD_TYPE_ZSUPER, 0, noex);
@@ -820,10 +846,11 @@ rb_method_boundp(VALUE klass, ID id, int ex)
     return 0;
 }
 
+extern ID rb_check_attr_id(ID id);
+
 void
 rb_attr(VALUE klass, ID id, int read, int write, int ex)
 {
-    const char *name;
     ID attriv;
     VALUE aname;
     rb_method_flag_t noex;
@@ -846,16 +873,11 @@ rb_attr(VALUE klass, ID id, int read, int write, int ex)
 	}
     }
 
-    if (!rb_is_local_id(id) && !rb_is_const_id(id)) {
-	rb_name_error(id, "invalid attribute name `%s'", rb_id2name(id));
-    }
-    name = rb_id2name(id);
-    if (!name) {
+    aname = rb_id2str(rb_check_attr_id(id));
+    if (NIL_P(aname)) {
 	rb_raise(rb_eArgError, "argument needs to be symbol or string");
     }
-    aname = rb_sprintf("@%s", name);
-    rb_enc_copy(aname, rb_id2str(id));
-    attriv = rb_intern_str(aname);
+    attriv = rb_intern_str(rb_sprintf("@%"PRIsVALUE, aname));
     if (read) {
 	rb_add_method(klass, id, VM_METHOD_TYPE_IVAR, (void *)attriv, noex);
     }
@@ -872,12 +894,6 @@ rb_undef(VALUE klass, ID id)
     if (NIL_P(klass)) {
 	rb_raise(rb_eTypeError, "no class to undef method");
     }
-    if (rb_vm_cbase() == rb_cObject && klass == rb_cObject) {
-	rb_secure(4);
-    }
-    if (rb_safe_level() >= 4 && !OBJ_UNTRUSTED(klass)) {
-	rb_raise(rb_eSecurityError, "Insecure: can't undef `%s'", rb_id2name(id));
-    }
     rb_frozen_class_p(klass);
     if (id == object_id || id == id__send__ || id == idInitialize) {
 	rb_warn("undefining `%s' may cause serious problems", rb_id2name(id));
@@ -886,7 +902,8 @@ rb_undef(VALUE klass, ID id)
     me = search_method(klass, id, 0);
 
     if (UNDEFINED_METHOD_ENTRY_P(me) ||
-	UNDEFINED_REFINED_METHOD_P(me->def)) {
+	(me->def->type == VM_METHOD_TYPE_REFINED &&
+	 UNDEFINED_METHOD_ENTRY_P(me->def->body.orig_me))) {
 	const char *s0 = " class";
 	VALUE c = klass;
 
@@ -913,11 +930,13 @@ rb_undef(VALUE klass, ID id)
 /*
  *  call-seq:
  *     undef_method(symbol)    -> self
+ *     undef_method(string)    -> self
  *
  *  Prevents the current class from responding to calls to the named
  *  method. Contrast this with <code>remove_method</code>, which deletes
  *  the method from the particular class; Ruby will still search
  *  superclasses and mixed-in modules for a possible receiver.
+ *  String arguments are converted to symbols.
  *
  *     class Parent
  *       def hello
@@ -971,10 +990,12 @@ rb_mod_undef_method(int argc, VALUE *argv, VALUE mod)
 /*
  *  call-seq:
  *     mod.method_defined?(symbol)    -> true or false
+ *     mod.method_defined?(string)    -> true or false
  *
  *  Returns +true+ if the named method is defined by
  *  _mod_ (or its included modules and, if _mod_ is a class,
  *  its ancestors). Public and protected methods are matched.
+ *  String arguments are converted to symbols.
  *
  *     module A
  *       def method1()  end
@@ -1013,7 +1034,7 @@ check_definition(VALUE mod, VALUE mid, rb_method_flag_t noex)
     const rb_method_entry_t *me;
     ID id = rb_check_id(&mid);
     if (!id) return Qfalse;
-    me = rb_method_entry_without_refinements(mod, id, 0);
+    me = rb_method_entry(mod, id, 0);
     if (me) {
 	if (VISI_CHECK(me->flag, noex))
 	    return Qtrue;
@@ -1024,10 +1045,12 @@ check_definition(VALUE mod, VALUE mid, rb_method_flag_t noex)
 /*
  *  call-seq:
  *     mod.public_method_defined?(symbol)   -> true or false
+ *     mod.public_method_defined?(string)   -> true or false
  *
  *  Returns +true+ if the named public method is defined by
  *  _mod_ (or its included modules and, if _mod_ is a class,
  *  its ancestors).
+ *  String arguments are converted to symbols.
  *
  *     module A
  *       def method1()  end
@@ -1056,10 +1079,12 @@ rb_mod_public_method_defined(VALUE mod, VALUE mid)
 /*
  *  call-seq:
  *     mod.private_method_defined?(symbol)    -> true or false
+ *     mod.private_method_defined?(string)    -> true or false
  *
  *  Returns +true+ if the named private method is defined by
  *  _ mod_ (or its included modules and, if _mod_ is a class,
  *  its ancestors).
+ *  String arguments are converted to symbols.
  *
  *     module A
  *       def method1()  end
@@ -1088,10 +1113,12 @@ rb_mod_private_method_defined(VALUE mod, VALUE mid)
 /*
  *  call-seq:
  *     mod.protected_method_defined?(symbol)   -> true or false
+ *     mod.protected_method_defined?(string)   -> true or false
  *
  *  Returns +true+ if the named protected method is defined
  *  by _mod_ (or its included modules and, if _mod_ is a
  *  class, its ancestors).
+ *  String arguments are converted to symbols.
  *
  *     module A
  *       def method1()  end
@@ -1218,15 +1245,11 @@ rb_alias(VALUE klass, ID name, ID def)
     }
 
     rb_frozen_class_p(klass);
-    if (klass == rb_cObject) {
-	rb_secure(4);
-    }
 
   again:
     orig_me = search_method(klass, def, &defined_class);
 
-    if (UNDEFINED_METHOD_ENTRY_P(orig_me) ||
-	UNDEFINED_REFINED_METHOD_P(orig_me->def)) {
+    if (UNDEFINED_METHOD_ENTRY_P(orig_me)) {
 	if ((!RB_TYPE_P(klass, T_MODULE)) ||
 	    (orig_me = search_method(rb_cObject, def, 0),
 	     UNDEFINED_METHOD_ENTRY_P(orig_me))) {
@@ -1240,7 +1263,7 @@ rb_alias(VALUE klass, ID name, ID def)
 	goto again;
     }
     if (RB_TYPE_P(defined_class, T_ICLASS)) {
-	VALUE real_class = RBASIC(defined_class)->klass;
+	VALUE real_class = RBASIC_CLASS(defined_class);
 	if (real_class && RCLASS_ORIGIN(real_class) == defined_class)
 	    defined_class = real_class;
     }
@@ -1282,23 +1305,15 @@ rb_mod_alias_method(VALUE mod, VALUE newname, VALUE oldname)
     return mod;
 }
 
-static void
-secure_visibility(VALUE self)
-{
-    if (rb_safe_level() >= 4 && !OBJ_UNTRUSTED(self)) {
-	rb_raise(rb_eSecurityError,
-		 "Insecure: can't change method visibility");
-    }
-}
-
 static void
 set_method_visibility(VALUE self, int argc, VALUE *argv, rb_method_flag_t ex)
 {
     int i;
-    secure_visibility(self);
 
     if (argc == 0) {
-	rb_warning("%s with no argument is just ignored", rb_id2name(rb_frame_callee()));
+	rb_warning("%"PRIsVALUE" with no argument is just ignored",
+		   QUOTE_ID(rb_frame_callee()));
+	return;
     }
 
     for (i = 0; i < argc; i++) {
@@ -1309,63 +1324,66 @@ set_method_visibility(VALUE self, int argc, VALUE *argv, rb_method_flag_t ex)
 	}
 	rb_export_method(self, id, ex);
     }
-    rb_clear_cache_by_class(self);
+}
+
+static VALUE
+set_visibility(int argc, VALUE *argv, VALUE module, rb_method_flag_t ex)
+{
+    if (argc == 0) {
+	SCOPE_SET(ex);
+    }
+    else {
+	set_method_visibility(module, argc, argv, ex);
+    }
+    return module;
 }
 
 /*
  *  call-seq:
  *     public                 -> self
  *     public(symbol, ...)    -> self
+ *     public(string, ...)    -> self
  *
  *  With no arguments, sets the default visibility for subsequently
  *  defined methods to public. With arguments, sets the named methods to
  *  have public visibility.
+ *  String arguments are converted to symbols.
  */
 
 static VALUE
 rb_mod_public(int argc, VALUE *argv, VALUE module)
 {
-    secure_visibility(module);
-    if (argc == 0) {
-	SCOPE_SET(NOEX_PUBLIC);
-    }
-    else {
-	set_method_visibility(module, argc, argv, NOEX_PUBLIC);
-    }
-    return module;
+    return set_visibility(argc, argv, module, NOEX_PUBLIC);
 }
 
 /*
  *  call-seq:
  *     protected                -> self
  *     protected(symbol, ...)   -> self
+ *     protected(string, ...)   -> self
  *
  *  With no arguments, sets the default visibility for subsequently
  *  defined methods to protected. With arguments, sets the named methods
  *  to have protected visibility.
+ *  String arguments are converted to symbols.
  */
 
 static VALUE
 rb_mod_protected(int argc, VALUE *argv, VALUE module)
 {
-    secure_visibility(module);
-    if (argc == 0) {
-	SCOPE_SET(NOEX_PROTECTED);
-    }
-    else {
-	set_method_visibility(module, argc, argv, NOEX_PROTECTED);
-    }
-    return module;
+    return set_visibility(argc, argv, module, NOEX_PROTECTED);
 }
 
 /*
  *  call-seq:
  *     private                 -> self
  *     private(symbol, ...)    -> self
+ *     private(string, ...)    -> self
  *
  *  With no arguments, sets the default visibility for subsequently
  *  defined methods to private. With arguments, sets the named methods
  *  to have private visibility.
+ *  String arguments are converted to symbols.
  *
  *     module Mod
  *       def a()  end
@@ -1380,21 +1398,17 @@ rb_mod_protected(int argc, VALUE *argv, VALUE module)
 static VALUE
 rb_mod_private(int argc, VALUE *argv, VALUE module)
 {
-    secure_visibility(module);
-    if (argc == 0) {
-	SCOPE_SET(NOEX_PRIVATE);
-    }
-    else {
-	set_method_visibility(module, argc, argv, NOEX_PRIVATE);
-    }
-    return module;
+    return set_visibility(argc, argv, module, NOEX_PRIVATE);
 }
 
 /*
  *  call-seq:
  *     mod.public_class_method(symbol, ...)    -> mod
+ *     mod.public_class_method(string, ...)    -> mod
  *
  *  Makes a list of existing class methods public.
+ *
+ *  String arguments are converted to symbols.
  */
 
 static VALUE
@@ -1407,10 +1421,13 @@ rb_mod_public_method(int argc, VALUE *argv, VALUE obj)
 /*
  *  call-seq:
  *     mod.private_class_method(symbol, ...)   -> mod
+ *     mod.private_class_method(string, ...)   -> mod
  *
  *  Makes existing class methods private. Often used to hide the default
  *  constructor <code>new</code>.
  *
+ *  String arguments are converted to symbols.
+ *
  *     class SimpleSingleton  # Not thread safe
  *       private_class_method :new
  *       def SimpleSingleton.create(*args, &block)
@@ -1431,10 +1448,13 @@ rb_mod_private_method(int argc, VALUE *argv, VALUE obj)
  *  call-seq:
  *     public
  *     public(symbol, ...)
+ *     public(string, ...)
  *
  *  With no arguments, sets the default visibility for subsequently
  *  defined methods to public. With arguments, sets the named methods to
  *  have public visibility.
+ *
+ *  String arguments are converted to symbols.
  */
 
 static VALUE
@@ -1443,6 +1463,18 @@ top_public(int argc, VALUE *argv)
     return rb_mod_public(argc, argv, rb_cObject);
 }
 
+/*
+ *  call-seq:
+ *     private
+ *     private(symbol, ...)
+ *     private(string, ...)
+ *
+ *  With no arguments, sets the default visibility for subsequently
+ *  defined methods to private. With arguments, sets the named methods to
+ *  have private visibility.
+ *
+ *  String arguments are converted to symbols.
+ */
 static VALUE
 top_private(int argc, VALUE *argv)
 {
@@ -1452,6 +1484,7 @@ top_private(int argc, VALUE *argv)
 /*
  *  call-seq:
  *     module_function(symbol, ...)    -> self
+ *     module_function(string, ...)    -> self
  *
  *  Creates module functions for the named methods. These functions may
  *  be called with the module as a receiver, and also become available
@@ -1460,6 +1493,7 @@ top_private(int argc, VALUE *argv)
  *  independently. The instance-method versions are made private. If
  *  used with no arguments, subsequently defined methods become module
  *  functions.
+ *  String arguments are converted to symbols.
  *
  *     module Mod
  *       def one
@@ -1496,7 +1530,6 @@ rb_mod_modfunc(int argc, VALUE *argv, VALUE module)
 	rb_raise(rb_eTypeError, "module_function must be called for modules");
     }
 
-    secure_visibility(module);
     if (argc == 0) {
 	SCOPE_SET(NOEX_MODFUNC);
 	return module;
@@ -1581,8 +1614,8 @@ rb_obj_respond_to(VALUE obj, ID id, int priv)
 			(FL_TEST(klass, FL_SINGLETON) ? '.' : '#'),
 			QUOTE_ID(id));
 		if (!NIL_P(location)) {
-		    VALUE path = RARRAY_PTR(location)[0];
-		    VALUE line = RARRAY_PTR(location)[1];
+		    VALUE path = RARRAY_AREF(location, 0);
+		    VALUE line = RARRAY_AREF(location, 1);
 		    if (!NIL_P(path)) {
 			rb_compile_warn(RSTRING_PTR(path), NUM2INT(line),
 					"respond_to? is defined here");
@@ -1604,6 +1637,7 @@ rb_respond_to(VALUE obj, ID id)
 /*
  *  call-seq:
  *     obj.respond_to?(symbol, include_all=false) -> true or false
+ *     obj.respond_to?(string, include_all=false) -> true or false
  *
  *  Returns +true+ if _obj_ responds to the given method.  Private and
  *  protected methods are included in the search only if the optional
@@ -1615,6 +1649,9 @@ rb_respond_to(VALUE obj, ID id)
  *
  *  If the method is not defined, <code>respond_to_missing?</code>
  *  method is called and the result is returned.
+ *
+ *  When the method name parameter is given as a string, the string is
+ *  converted to a symbol.
  */
 
 static VALUE
@@ -1641,13 +1678,17 @@ obj_respond_to(int argc, VALUE *argv, VALUE obj)
 /*
  *  call-seq:
  *     obj.respond_to_missing?(symbol, include_all) -> true or false
+ *     obj.respond_to_missing?(string, include_all) -> true or false
  *
  *  DO NOT USE THIS DIRECTLY.
  *
  *  Hook method to return whether the _obj_ can respond to _id_ method
  *  or not.
  *
- *  See #respond_to?.
+ *  When the method name parameter is given as a string, the string is
+ *  converted to a symbol.
+ *
+ *  See #respond_to?, and the example of BasicObject.
  */
 static VALUE
 obj_respond_to_missing(VALUE obj, VALUE mid, VALUE priv)
@@ -1684,15 +1725,6 @@ Init_eval_method(void)
     rb_define_private_method(rb_singleton_class(rb_vm_top_self()),
 			     "private", top_private, -1);
 
-    object_id = rb_intern("object_id");
-    added = rb_intern("method_added");
-    singleton_added = rb_intern("singleton_method_added");
-    removed = rb_intern("method_removed");
-    singleton_removed = rb_intern("singleton_method_removed");
-    undefined = rb_intern("method_undefined");
-    singleton_undefined = rb_intern("singleton_method_undefined");
-    attached = rb_intern("__attached__");
-
     {
 #define REPLICATE_METHOD(klass, id, noex) \
 	rb_method_entry_set((klass), (id), \
